
# occiput 
# Stefano Pedemonte
# Aalto University, School of Science, Helsinki
# Oct 2014, Helsinki 
# Harvard University, Martinos Center for Biomedical Imaging 
# Boston, MA, USA
# April 2014
# Nov. 2015 


__all__ = ["guess_file_type_by_name"] 


def guess_file_type_by_name(filename): 
    if filename.endswith("h5"): 
        return "h5"
    elif filename.endswith(".l.hdr"): 
        return "interfile_listmode_header"
    if filename.endswith(".l"): 
        return "interfile_listmode_data"
    if filename.endswith(".h33") or filename.endswith(".s.hdr"): 
        return "interfile_projection_header"
    elif filename.endswith(".a") or filename.endswith(".s"):
        return "interfile_projection_data"
    elif filename.endswith(".v.hdr"): 
        return "interfile_volume_header"
    elif filename.endswith(".v"): 
        return "interfile_volume_data"
    elif filename.endswith(".nii.gz"): 
        return "nifti_compressed"
    elif filename.endswith(".nii"): 
        return "nifti"
    elif filename.endswith(".mat"): 
        return "mat"
    else: 
        return "unknown"
        









class BaseCollimatorSPECT(): 
    def __init__(self): 
        self._name = "Generic Collimator" 
        self._collimator_type = "Generic Collimator" 
        self._manufacturer = "No manufacturer"
        self._p_simulation_method    = "analytical" 
        self._p_hole_diameter_mm     = 1.0 
        self._p_thickness_mm         = 40.0 
        self._p_gap_scintillatpr_mm  = 0.0 
        self._p_septum_mm            = 0.3 
        self._p_hole_shape           = "hexagonal" 
        self._need_to_recompute_psf = True 
        self._previous_energy_request = 0
        self._psf = None

    def get_hole_diameter(self):
        return self._p_hole_diameter_mm

    def set_hole_diameter(self,diameter_mm):
        if not np.isscalar(diameter_mm): 
            raise BadParameter('Expected a scalar value.')
        self._p_hole_diameter = diameter_mm
        self._need_to_recompute_psf = True

    def get_thickness(self):
        return self._p_thickness_mm

    def set_thickness(self,thickness_mm):
        if not np.isscalar(thickness_mm): 
            raise BadParameter('Expected a scalar value.')
        self._p_thickness_mm = thickness_mm
        self._need_to_recompute_psf = True

    def get_gap_scintillator(self):
        return self._p_gap_scintillator_mm

    def set_gap_scintillator(self,gap_scintillator_mm):
        if not np.isscalar(gap_scintillator_mm): 
            raise BadParameter('Expected a scalar value.')
        self._p_gap_scintillator_mm = gap_scintillator_mm
        self._need_to_recompute_psf = True

    def get_septum(self):
        return self._p_septum_mm

    def set_septum(self,septum_mm):
        if not np.isscalar(septum_mm): 
            raise BadParameter('Expected a scalar value.')
        self._p_septum_mm = septum_mm
        self._need_to_recompute_psf = True

    def get_hole_shape(self):
        return self._p_hole_shape

    def set_hole_shape(self,hole_shape):
        if str(hole_shape) == "hexagonal" or str(hole_shape) == "circular": 
            self._p_hole_shape = str(hole_shape)
        else: 
            raise BadParameter('Unknown hole shape') 
        self._need_to_recompute_psf = True

    def get_simulation_method(self): 
        return self._p_simulation_method 

    def set_simulation_method(self,simulation_method): 
        if str(simulation_method) == "analytical": 
            self._p_simulation_method = str(simulation_method)
        else: 
            raise BadParameter('Unknown simulation method') 
        self._need_to_recompute_psf = True

    def get_parameters(self):
        parameters = {}
        dic = self.__dict__
        for k in dic.keys(): 
            if k.startswith('_p_'):
                parameters[k[3:]]=dic[k]        
        return parameters 

    def get_psf(self,energy_kev): 
        if self._need_to_recompute_psf or energy_kev!=self._previous_energy_request: 
            self._compute_psf(energy_kev)
            self._previous_energy_request = energy_kev
        return self._psf 

    def _compute_psf(self,energy_kev): 
        print_medium_verbose("Recomputing PSF collimator ..")
        self._need_to_recompute_psf = False
        self._psf = 0



class LEHR(BaseCollimatorSPECT):
    def __init__(self): 
        BaseCollimatorSPECT.__init__(self)
        self._name = "Generic LEHR Collimator"
        self._collimator_type = "Low Energy High Resolution - LEHR"
        self._manufacturer = "No manufacturer" 

    def get_psf(self,energy_kev): 
        if self._need_to_recompute_psf or energy_kev!=self._previous_energy_request: 
            self._compute_psf(energy_kev)
            self._previous_energy_request = energy_kev
        return self._psf 

    def _compute_psf(self,energy_kev): 
        print_medium_verbose("Recomputing PSF collimator ..")
        self._need_to_recompute_psf = False
        self._psf = 0


class HELR(BaseCollimatorSPECT):
    def __init__(self): 
        BaseCollimatorSPECT.__init__(self)
        self._name = "Generic HELR Collimator"
        self._collimator_type = "High Energy Low Resolution - HELR"
        self._manufacturer = "No manufacturer" 

    def get_psf(self,energy_kev): 
        if self._need_to_recompute_psf or energy_kev!=self._previous_energy_request: 
            self._compute_psf(energy_kev)
            self._previous_energy_request = energy_kev
        return self._psf 

    def _compute_psf(self,energy_kev): 
        print 1
        print_low_verbose("Recomputing PSF collimator ..")
        self._need_to_recompute_psf = False
        self._psf = 0







# occiput 
# Stefano Pedemonte
# Aalto University, School of Science, Helsinki
# Oct 2014, Helsinki 
# Harvard University, Martinos Center for Biomedical Imaging 
# Boston, MA, USA
# March 2015


from occiput.DataSources.FileSources.Volume import import_nifti

def import_kspace(filename): 
    return import_nifti(filename) 





# occiput 
# Stefano Pedemonte 
# April 2014 
# Harvard University, Martinos Center for Biomedical Imaging 
# Boston, MA, USA 
# Nov. 2015 


__all__ = ['import_interfile_projection','export_interfile_projection','import_interfile_volume',
           'export_interfile_volume','guess_file_type_by_name','import_nifti','import_PET_Projection',
           'import_dicom','import_mask','import_dicom_series','load_freesurfer_lut_file',
           'load_vnav_mprage','import_listmode','download_Dropbox','import_kspace','load_motion_sensor_data',
           'Brain_PET_Physiology', 'Biograph_mMR_Physiology'] 



from Volume import import_nifti
from Volume import import_dicom
from Volume import import_mask
from Volume import import_dicom_series

from PET_listmode import import_listmode, convert_listmode_dicom_to_interfile
from PET_projection import import_interfile_projection, export_interfile_projection, import_PET_Projection 
from PET_volume import import_interfile_volume, export_interfile_volume
from MR_kspace import import_kspace

from LookupTable import load_freesurfer_lut_file

from Web import download_Dropbox

from vNAV import load_vnav_mprage
from MR_motion_sensors import load_motion_sensor_data

from Physiology import Brain_PET_Physiology, Biograph_mMR_Physiology

from Files import guess_file_type_by_name






# occiput 
# Stefano Pedemonte
# Aalto University, School of Science, Helsinki
# Oct 2013, Helsinki 
# Martinos Center for Biomedical Imaging, Harvard University/MGH, Boston
# Dec. 2013, Boston
# Martinos Center for Biomedical Imaging, Harvard University/MGH, Boston
# Jan. 2015, Boston
# Feb. 2015, Helsinki
# Nov. 2015, Boston 


# Interface to list-mode data. In order to cope with various data formats of the list-mode data and 
# to enable fast decoding, occiput uses a simple plug-in system. The routines 
# to decode the list-mode data are defined in a C function that is loaded as a plugin. 
# The simple plugin system handles the compilation of the code. 



from occiput.Reconstruction.PET.PET_Projection import PET_Projection, PET_Projection_Sparsity
from occiput.Core.NiftyPy_wrap import PET_listmode_get_measurement_static, PET_listmode_get_measurement

__all__ = ["Listmode_Loader"]


class ListmodeWrapper(): 
    def open_stream(self, stream): 
        PET_listmode_open_stream(stream)
    
    def close_stream(self): 
        PET_listmode_close_stream()
    
    def load_scanner_plugin(self, plugin): 
        scanner_library = plugin.get_library_path()
        PET_listmode_load_scanner_plugin(scanner_library) 

    def load_stream_plugin(self, plugin): 
        stream_library = plugin.get_library_path()
        PET_listmode_load_stream_plugin(stream_library)     
    
    def bin_events(self, n_events, Naxial, Nazim, Nu, Nv, time_bins): 
        PET_listmode_bin(n_events, Naxial, Nazim, Nu, Nv, time_bins)

    def bin_events_intrinsic(self, n_events): 
        PET_listmode_bin_intrinsic(n_events)

    def get_static_measurement(self): 
        R = PET_listmode_get_static()
        return R['counts'], R['offsets'], R['locations'], R['time_start'], R['time_end'] 

    def get_dynamic_measurement(self): 
        R = PET_listmode_get_dynamic()
        return R['counts'], R['offsets'], R['locations'], R['time_start'], R['time_end'] 



class ListmodeLoader(): 
    def __init__(self): 
        self._LM = ListmodeWrapper()
    
    def set_streamer(self, plugin): 
        # call C function that loads dynamically the open(), close() and read() functions from the streamer dynamic library
        self._LM.load_stream_plugin( plugin )
    
    def set_scanner(self, plugin): 
        # call C function that loads dynamically the index_to_lor() function of the scanner plugin 
        self._LM.load_scanner( plugin ) 
    
    def open_stream(self, stream): 
        # call the C function that opens a list-mode stream 
        self._LM.open_stream( stream )
    
    def close_stream(self): 
        # call the C function that closes the list-mode stream 
        self._LM.close_stream()

    def make_projection_given_binning(self, binning, n_events=-1, time_bins=None):  
        # call the C function that parses the list-mode data and bins it into (dynamic) projection data 
        self._LM.bin_events(n_events, binning.N_axial, binning.N_azimuthal, binning.N_u, binning.N_v, time_bins)
        counts,offsets,locations,time_start,time_end = self._LM.get_static_measurement()
        time_bins = int32(linspace(time_start, time_end, 2))
        return PET_Projection( binning, counts, offsets, locations, time_bins) 

    def make_projection_intrinsic_geometry(self, n_events=-1, time_bins=None, N_azimuthal=-1): 
        # call the C function that parses the list-mode data and bins it into (dynamic) projection data 
        self._LM.bin_events_intrinsic(n_events, binning.N_axial, binning.N_azimuthal, binning.N_u, binning.N_v, time_bins)
        counts,offsets,locations,time_start,time_end = self._LM.get_static_measurement()
        time_bins = int32(linspace(time_start, time_end, 2))
        return PET_Projection( binning, counts, offsets, locations, time_bins) 






# occiput 
# Stefano Pedemonte
# Aalto University, School of Science, Helsinki
# Oct 2013, Helsinki 
# Martinos Center for Biomedical Imaging, Harvard University/MGH, Boston
# Dec. 2013, Boston
# Martinos Center for Biomedical Imaging, Harvard University/MGH, Boston
# Jan. 2015, Boston
# Feb. 2015, Helsinki
# Nov. 2015, Boston 


# This file defines a PET projection: the data structure that contains projection data. 
# In occiput.io, the projection data is of two types: sparse or non-sparse. 
# The sparse projection data is memory efficient and leads to efficient projection and 
# back-projection. Low memory is necessary in order to use hundreds or thousands of sinograms 
# for dynamic and kinetic imaging. There is a single object to represent projection data, 
# with a flag that sets the type: sparse or non-sparse. The ray-tracer knows how to handle the projection data object.  


import occiput 
from occiput.Visualization import ipy_table, has_ipy_table, svgwrite, has_svgwrite 
from occiput.Core.Print import rad_to_deg, deg_to_rad, array_to_string
from occiput.Core.NiftyPy_wrap import PET_compress_projection, PET_uncompress_projection, PET_initialize_compression_structure #, PET_compress_projection_array 

import h5py 
import copy
import os
import inspect 
from numpy import isscalar, linspace, int32, uint32, uint16, ones, zeros, pi, sqrt, float32, float64, where, ndarray, nan, tile
from numpy import inf, asarray, concatenate, fromfile, maximum, exp, asfortranarray, fliplr, transpose 
from numpy.random import poisson



__all__ = ["DEFAULT_BINNING","Binning", "PET_Projection_Sparsity", "PET_Projection",
          "PET_compress_projection", "PET_uncompress_projection", "PET_initialize_compression_structure","display_PET_Projection_geometry",] #"PET_compress_projection_array"] 

    
def display_PET_Projection_geometry(): 
    filename = inspect.getfile(occiput).strip("__init__.pyc")+"Data"+os.sep+"occiput_ray_tracer_PET.png"
    from IPython.display import Image
    img = Image(filename=filename, width=1000) 
    return img 



EPS = 1e-10
DEFAULT_BINNING = {      "n_axial":                120,
                         "n_azimuthal":            5,
                         "angles_axial":           float32( linspace(0,pi-pi/120.0,120) ), 
                         "angles_azimuthal":       float32( [-0.5, -0.25, 0.0, 0.25, 0.5] ),
                         "size_u":                 256.0,
                         "size_v":                 256.00,
                         "n_u":                    128,
                         "n_v":                    128,   }



class Binning(): 
    """PET detectors binning. """
    def __init__(self,parameters=None): 
        self._initialised = 0
        self.name = "Unknown binning name"
        if parameters is None: 
            self.load_from_dictionary(DEFAULT_BINNING)  
        elif type(parameters) == dict:  
            self.load_from_dictionary(parameters)        
        else: 
            raise UnknownParameter('Parameter %s specified for the construction of Binning is not compatible. '%str(parameters)) 
            
    def load_from_dictionary(self,dictionary):
        self.N_axial                = dictionary['n_axial']                    # Number of axial bins 
        self.N_azimuthal            = dictionary['n_azimuthal']                # Number of azimuthal bins
        self.angles_axial           = dictionary['angles_axial']               # Axial angles radians
        self.angles_azimuthal       = dictionary['angles_azimuthal']           # Azimuthal angles radians
        self.size_u                 = dictionary['size_u']                     # Size of the detector plane, axis u, [adimensional]
        self.size_v                 = dictionary['size_v']                     # Size of the detector plane, axis v,  [adimensional]
        self.N_u                    = dictionary['n_u']                        # Number of pixels of the detector plan, axis u 
        self.N_v                    = dictionary['n_v']                        # Number of pixels of the detector plan, axis v 
        self.angles_axial           = dictionary['angles_axial']    
        
    def get_angles(self, subsets_matrix=None): 
        angles = zeros((2,self.N_azimuthal,self.N_axial))
        angles[0,:,:] = tile(self.angles_axial,(self.N_azimuthal,1))
        angles[1,:,:] = tile(self.angles_azimuthal,(self.N_axial,1)).T
        
        if subsets_matrix is not None: 
            subsets_matrix = int32(subsets_matrix)
            N = subsets_matrix.sum()
            angles2 = angles.reshape(2,self.N_azimuthal*self.N_axial).copy()
            angles = zeros([2, 1, N])
            angles[0,0,:] = ( angles2[0,subsets_matrix.flatten()==1] ).reshape([N,])
            angles[1,0,:] = ( angles2[1,subsets_matrix.flatten()==1] ).reshape([N,])
        return float32(angles)

    def __repr__(self): 
        s = "PET Binning: \n"        
        s = s+" - N_axial_bins:             %d \n"%self.N_axial 
        s = s+" - N_azimuthal_bins:         %d \n"%self.N_azimuthal 
        s = s+" - Size_u:                   %f \n"%self.size_u
        s = s+" - Size_v:                   %f \n"%self.size_v        
        s = s+" - N_u:                      %d \n"%self.N_u
        s = s+" - N_v:                      %d \n"%self.N_v
        s = s+" - Angles axial:             %s \n"%array_to_string(self.angles_axial)
        s = s+" - Angles azimuthal:         %s \n"%array_to_string(self.angles_azimuthal)
        return s

    def _repr_html_(self):       #FIXME: add angles_axial and angles_azimuthal 
        if not has_ipy_table: 
            return "Please install ipy_table."
        table_data = [['N_axial',self.N_axial],['N_azimuthal',self.N_azimuthal],['Size_u',self.size_u],['Size_v',self.size_v],['N_u',self.N_u],['N_v',self.N_v]] 
        N_per_row = 12
        for i in range(len(self.angles_azimuthal)/N_per_row+1): 
            i_start = i*N_per_row
            i_end = (i+1)*N_per_row 
            if i_end>=len(self.angles_azimuthal): 
                i_end=len(self.angles_azimuthal)
            if i==0: 
                table_data.append([ "Angles_azimuthal [deg]", array_to_string(rad_to_deg(self.angles_azimuthal[i_start:i_end]),"%3.4f ") ]) 
            else: 
                table_data.append([ "", array_to_string(rad_to_deg(self.angles_azimuthal[i_start:i_end]),"%3.4f ") ])
        i_start=0
        i_end  =len(self.angles_axial)-1
        if i_end-i_start >= 4: 
            table_data.append([ "Angles_axial [deg]", str(rad_to_deg(self.angles_axial[i_start]))+" "+str(rad_to_deg(self.angles_axial[i_start+1]))+"  ...  "+str(rad_to_deg(self.angles_axial[i_end-1]))+"   "+str(rad_to_deg(self.angles_axial[i_end])), ])
        table = ipy_table.make_table(table_data)
        table = ipy_table.apply_theme('basic_left')
        #table = ipy_table.set_column_style(0, color='lightBlue')
        table = ipy_table.set_global_style(float_format="%3.3f")        
        return table._repr_html_()
        
    def _make_svg(self): 
        if not has_svgwrite: 
            self._svg_string = None
            return self._svg_string

        w = '100%'
        h = '100%'
        
        dwg = svgwrite.Drawing('test.svg',size=(w,h), profile='full', debug=True)
        dwg.viewbox(width=100, height=100)
        
        # IMAGING VOLUME
        rect = dwg.add(dwg.rect(insert=(30, 30), size=(40, 40), rx=0.5, ry=0.5))
        rect.fill('grey',opacity=0.02).stroke('grey',width=0.3,opacity=0.02)
        
        # GEOMETRIC NOTATIONS 
        # circle, gantry rotation 
        circle = dwg.add(dwg.circle(center=(50, 50), r=30))
        circle.fill('none').stroke('grey', width=0.1).dasharray([0.5, 0.5]) 
        # center
        circle = dwg.add(dwg.circle(center=(50, 50), r=0.5))
        circle.fill('grey',opacity=0.1).stroke('grey', width=0.1)    
        line = dwg.add(dwg.line(start=(50-1,50), end=(50+1,50)))
        line.stroke('grey', width=0.1) 
        line = dwg.add(dwg.line(start=(50,50-1), end=(50,50+1)))
        line.stroke('grey', width=0.1) 
        
        #line = dwg.add(dwg.polyline([(10, 10), (10, 100), (100, 100), (100, 10), (10, 10)],stroke='black', fill='none'))
        self._svg_string = dwg.tostring() 
        return self._svg_string

    def _repr_svg_(self): 
        self._make_svg()
        return self._svg_string    

    def display_geometry(self):
        return display_PET_Projection_geometry()
        


class PET_Projection_Sparsity(): 
    """Represents the sparsity pattern of projection data and implements methods to compress and uncompress data according to the sparsity pattern of self. 
    If offsets and locations are not specified as parameters of the constructor, the data structure represents dense projection data. """
    def __init__(self, N_axial, N_azimuthal, N_u, N_v , offsets=None, locations=None): 
        if offsets is None or locations is None: 
            # Uncompressed 
            [offsets,locations] = PET_initialize_compression_structure(N_axial,N_azimuthal,N_u,N_v) 
            offsets = offsets.transpose().reshape((N_azimuthal, N_axial))
        self.offsets     = offsets
        self.locations   = locations
        self.N_axial     = int32(N_axial)
        self.N_azimuthal = int32(N_azimuthal)
        self.N_u         = N_u
        self.N_v         = N_v

    def compress_data(self, data): 
        """Compresses an array of uncompressed projection data, according to the sparsity pattern of self (self.offsets and self.locations). """
        compressed_data = PET_compress_projection(self.offsets, data, self.locations, self.N_u, self.N_v)
        #print "Compression done"
        return compressed_data

    def uncompress_data(self, data): 
        """Uncompresses a data array that contains projection compressed according to the sparsity pattern of self (self.offsets and self.locations). """
        uncompressed_data = PET_uncompress_projection(self.offsets, data, self.locations, self.N_u, self.N_v)
        uncompressed_data.reshape([self.offsets.shape[1], self.offsets.shape[0], self.N_u, self.N_v]) 
        #print "Uncompression done"
        return uncompressed_data

    def same_sparsity_as(self, obj): 
        """Returns True if the argument is an object of type PET_Projection_Sparsity with the same identical sparsity pattern as self. """
        if not hasattr(obj,'offsets'): 
            return False 
        if not hasattr(obj,'locations'): 
            return False 
        return obj.offsets == self.offsets and obj.locations == self.locations

    def is_compressed(self): 
        return not self.is_uncompressed() 

    def is_uncompressed(self): 
        """Returns True if the sparsity pattern represents an uncompressed sampled projection. """
        return self.get_N_locations() == self.N_u * self.N_v * self.N_axial * self.N_azimuthal

    def get_N_locations(self): 
        """Returns the number of non-zero entries of the sparse projection data. """
        return self.locations.shape[1]

    # Overload math operators 
    def __and__(self, other): 
        """Overload the AND operator. """
        if not isinstance(other,self.__class__): 
            raise Exception("Binary operation must be with another object of the same type - PET_Projection_Sparsity. ")
        print "This is not implemented, it should be implemented at low level. "

    def __xor__(self, other): 
        if not isinstance(other,self.__class__): 
            raise Exception("Binary operation must be with another object of the same type - PET_Projection_Sparsity. ")
        print "This is not implemented, it should be implemented at low level. "
    
    def __or__(self, other): 
        if not isinstance(other,self.__class__): 
            raise Exception("Binary operation must be with another object of the same type - PET_Projection_Sparsity. ")
        print "This is not implemented, it should be implemented at low level. "

    def get_locations_per_plane(self):
        """Returns the number of active locations per projection plane."""
        shape = self.offsets.shape
        offsets = self.offsets.transpose().flatten()
        v = offsets.copy()
        v[0:-1] = offsets[1::]
        v[-1]   = self.locations.shape[1]
        sizes = v - offsets 
        return sizes.reshape(shape)

    def get_subset(self, subset_matrix):
        subset_matrix = int32(subset_matrix)
        indexes = subset_matrix == 1
        N = int32(subset_matrix.sum())
        locations = self.locations
        offsets   = self.offsets
        N_axial = self.N_axial
        N_azimuthal = self.N_azimuthal
        N_u = self.N_u
        N_v = self.N_v

        offsets_return = zeros((1,N), dtype=int32, order="F")
        sizes = self.get_locations_per_plane().reshape(indexes.shape)
        
        N_locations = sizes[indexes].sum()
        locations_return = zeros((3,N_locations), dtype=uint16, order="F")
    
        ii = 0
        previous_offset = 0
        for iax in range(N_axial):
            for iaz in range(N_azimuthal): 
                if subset_matrix[iaz,iax]: 
                    ii = ii + 1
                    size = sizes[iaz,iax]
                    if ii<N: 
                        offsets_return[0,ii] = offsets_return[0,ii-1]+size
                    locations_return[:,offsets_return[0,ii-1]:offsets_return[0,ii-1]+size] = locations[:,offsets[iaz,iax]:offsets[iaz,iax] + size]
        return PET_Projection_Sparsity(N, 1, N_u, N_v , offsets = offsets_return, locations = locations_return)

    def display_geometry(self):
        return display_PET_Projection_geometry()


class PET_Projection(): 
    """Sparse PET projection object. If offsets=None or locations=None in the constructor, the projection is uncompressed. """
    def __init__(self, binning, data=None, offsets=None, locations=None, time_bins=None, subsets_matrix=None): 
        """Sparse PET projection object. If offsets=None or locations=None in the constructor, the projection is uncompressed. """
        # The data structure that represents a projection is composed of: 
        #1) Sparsity structure 
        #2) Time binning (list of time flags)
        #3) Size of projection planes - in units of [mm]
        #4) Angular steps in axial and azimuthal directions - in units of [radians] 
        #5) The projection data (e.g. counts)
        self.binning = binning 
        self.angles = self.binning.get_angles(subsets_matrix=subsets_matrix)
        
        sparsity = PET_Projection_Sparsity(binning.N_axial, binning.N_azimuthal, binning.N_u, binning.N_v, offsets, locations)
        if subsets_matrix is not None: 
            self.sparsity = sparsity.get_subset(subsets_matrix)
        else: 
            self.sparsity = sparsity

        if time_bins is None: 
            time_bins = int32([0,1000])   # 1 second if not specified 
        self.time_bins = time_bins
        
        N_axial = int32(self.angles.shape[2])
        N_azimuthal = int32(self.angles.shape[1])
        
        if data is None: 
            data = zeros((N_axial, N_azimuthal, binning.N_u, binning.N_v))
        elif isscalar(data): 
            data = data + zeros((N_axial, N_azimuthal, binning.N_u, binning.N_v))
        if self.sparsity.is_uncompressed(): 
            data = data.reshape((N_axial, N_azimuthal, binning.N_u, binning.N_v))
        self.data = data     
          
    def get_binning(self): 
        """Returns the binning parameters (that is an instance of Binning). """
        return self.binning 

    def get_time_bins(self):
        """Returns the time bins (list of N+1 time flags in case of N time bins) associated to the projection. """
        return self.time_bins 
  
    def get_sparsity(self): 
        """Returns an instance of PET_Projection_Sparsity, which describes the sparsity pattern of the projection data. """
        return self.sparsity 
    
    def get_data(self):
        """Returns the (compressed) raw projection data (note that is can be accessed also as self.data ). """
        return self.data

    def get_angles(self): 
        return self.angles

    def compress_as_self(self, projection):
        """Returns a PET_Projection object obtained by sampling the PET_Projection object given as argument 
        according to the sparsity of self. The argument may be a compressed or uncompressed PET_Projection object. """
        # FIXME: take a PET_Proejction_sparsity as argument optionally
        # FIXME: verify that the two PET_Projection objects have the same binning 
        if isinstance(projection, PET_Projection): 
            if not projection.is_uncompressed(): 
                print "Re-compressing projection data, possible data loss. "
                uncompressed_projection = projection.uncompress_self() 
                uncompressed_data = uncompressed_projection.data
            else: 
                uncompressed_projection = projection
                uncompressed_data = uncompressed_projection.data
        else: 
            uncompressed_data = projection
            # FIXME: verify that size of the array is correct for offsets and locations 
        compressed_data = self.sparsity.compress_data(uncompressed_data) 
        return PET_Projection( self.get_binning(), compressed_data, self.sparsity.offsets, self.sparsity.locations, self.get_time_bins() )

    def compress_as(self, projection):
        """Returns a PET_Projection object obtained by sampling the projection data of self 
        according to the sparsity of the PET_Proejction object given as argument. 
        The argument may be a compressed or uncompressed PET_Projection object. """
        # FIXME: take a PET_Proejction_sparsity as argument optionally
        # FIXME: verify that the two PET_Projection objects have the same binning 
        uncompressed_data = self.sparsity.uncompress_data(self.data) 
        data = projection.sparsity.compress_data(uncompressed_data)
        return PET_Projection( projection.get_binning(), data, projection.sparsity.offsets, projection.sparsity.locations, projection.get_time_bins() )
        
    def compress_self(self): 
        if self.is_compressed(): 
            return self
        else: 
            print "Find the zeros and compress. "
            print "Not implemented. Please implement me. This needs low level implementation. Right now sparsity comes only from listmode data. "
            #offsets, locations, data = PET_compress_projection_array(self.data, self.N_axial, self.N_azimuthal, self.N_u, self.N_v)
    #return PET_Projection( self.get_binning(), data, offsets, locations, self.get_time_bins() )

    def uncompress_self(self): 
        """Returns an instance of PET_Projection obtained by zero-filling self. The projection object returned is not sparse. """
        if self.is_uncompressed(): 
            return self
        uncompressed_data = self.sparsity.uncompress_data(self.data) 
        return PET_Projection( self.get_binning(), uncompressed_data, offsets=None, locations=None, time_bins=self.get_time_bins() )

    def save_to_file(self, filename): 
        h5f = h5py.File(filename, 'w')
        h5f.create_dataset('offsets', data=self.sparsity.offsets)
        h5f.create_dataset('locations', data=self.sparsity.locations)
        h5f.create_dataset('data', data=self.data)
        h5f.create_dataset('time_bins', data=self.time_bins)
        binning = self.get_binning() 
        h5f.create_dataset('n_axial', data=binning.N_axial)
        h5f.create_dataset('n_azimuthal', data=binning.N_azimuthal)
        h5f.create_dataset('angles_axial', data=binning.angles_axial)
        h5f.create_dataset('angles_azimuthal', data=binning.angles_azimuthal)
        h5f.create_dataset('size_u', data=binning.size_u)
        h5f.create_dataset('size_v', data=binning.size_v)
        h5f.create_dataset('n_u', data=binning.N_u)
        h5f.create_dataset('n_v', data=binning.N_v)
        h5f.close()

    def same_sparsity_as(self, obj): 
        """Returns True if the argument is an object of type PET_Projection_Sparsity with the same identical sparsity pattern as self. """
        if not hasattr(obj,'sparsity'): 
            print "PET.py - same_sparsity(): Object of wrong type"  # FIXME:raise input error
        return self.sparsity.same_sparsity_as(obj.sparsity) 

    def is_compressed(self): 
        return not self.is_uncompressed() 

    def is_uncompressed(self): 
        """Returns True if the sparsity pattern represents an uncompressed projection. """
        return self.sparsity.is_uncompressed() 

    def get_N_locations(self): 
        """Returns the number of non-zero entries of the sparse projection data. """
        return self.sparsity.get_N_locations() 

    def max(self):
        return self.data.max() 
    
    def min(self):
        return self.data.min()

    def get_time_start(self):
        return self.time_bins[0]
    
    def get_time_end(self): 
        return self.time_bins[-1] 

    def get_duration(self): 
        return self.get_time_end() - self.get_time_start() 

    def get_integral(self): 
        return self.data.sum() 

    def get_compression_ratio(self):
        return (1.0 * self.get_N_locations()) / (self.sparsity.N_u*self.sparsity.N_v*self.sparsity.N_axial*self.sparsity.N_azimuthal)
        
    def get_listmode_loss(self): 
        return (1.0 * self.get_integral()) / self.get_N_locations()

    def get_projection_info(self):
        """Returns information related to the compression, such as the compression ratio. It returns a dictionary. """
        return self.sparsity.get_projection_info() 

    def to_nd_array(self): 
        """Uncompress (if compressed) and convert into an nd array. """
        if not self.is_uncompressed(): 
            uncompressed = self.uncompress_self() 
        else: 
            uncompressed = self 
        shape = (uncompressed.sparsity.N_axial, uncompressed.sparsity.N_azimuthal, uncompressed.sparsity.N_u, uncompressed.sparsity.N_v ) 
        return uncompressed.data.reshape(shape)

    def _is_3D_data(self): 
        if self.sparsity.N_axial == 1 or self.sparsity.N_azimuthal == 1: 
            return True 
        return False

    def _is_4D_data(self): 
        if self.sparsity.N_axial != 1 and self.sparsity.N_azimuthal != 1: 
            return True 
        return False

    def to_image(self,data,index_axial=0,index_azimuthal=0,scale=None,absolute_scale=False): 
        from PIL import Image
        a = float32(data[index_axial,index_azimuthal,:,:].reshape((data.shape[2],data.shape[3])))   
        if scale is None:
            a = 255.0*(a)/(a.max()+1e-12)
        else: 
            if absolute_scale: 
                a = scale*(a) 
            else: 
                a = scale*255.0*(a)/(a.max()+1e-12)
        im = Image.fromarray(a).convert("RGB")
        return im.rotate(90) 
        
    def display_in_browser(self,axial=True,azimuthal=False,index=0,scale=None): 
        self.display(axial=axial,azimuthal=azimuthal,index=index,scale=scale,open_browser=True)

    def display(self,axial=True,azimuthal=None,index=0,scale=None,open_browser=False): 
        """If 'axial' is set to True, then it displays projections along the axial direction, with azimuthal index given by 'index'. 
        If 'axial' is set to False, then it displays projections along the azimuthal direction, with axial index given by 'index'. """
        data = self.to_nd_array() 
        binning = self.get_binning()
        d = DisplayNode()
        images = []
        progress_bar = ProgressBar(height='6px', width='100%%', background_color=LIGHT_GRAY, foreground_color=GRAY) 
        if scale is not None:
            scale = scale*255.0/(data.max()+1e-12)
        else: 
            scale = 255.0/(data.max()+1e-12) 
        if axial is None and azimuthal is None:   # if axial and azimuthal are not specified, select them automatically based on data shape 
            axial = (self.sparsity.N_axial > 1) 
            azimuthal = (self.sparsity.N_azimuthal > 1)
        else: 
            if axial == None: 
                axial = False
            if azimuthal == None: 
                azimuthal = False 
        if axial and not azimuthal: 
            for i in range( self.sparsity.N_axial ): 
                images.append( self.to_image(data,i,index,scale=scale,absolute_scale=True) ) 
                progress_bar.set_percentage(i*100.0/self.sparsity.N_axial) 
        elif azimuthal and not axial: 
            for i in range( binning.N_azimuthal ): 
                images.append( self.to_image(data,index,i,scale=scale,absolute_scale=True) )
                progress_bar.set_percentage(i*100.0/binning.N_azimuthal)     
        else: 
            for i in range(self.sparsity.N_axial): 
                images_az = []
                for j in range(binning.N_azimuthal): 
                    images_az.append( self.to_image(data,i,j,scale=scale,absolute_scale=True) )
                images.append(images_az) 
                progress_bar.set_percentage(i*100.0/self.sparsity.N_axial)                         
        progress_bar.set_percentage(100.0) 
        return d.display('tipix', images, open_browser) 

    def _repr_html_(self): 
        return self.display()._repr_html_() 


    def _is_same_sparsity_(self, other): 
        if isinstance(other,self.__class__): 
            # FIXME: make sure that the two projections are compatible: same geometry and sparsity pattern
            return True 
        return False 

    def copy(self): 
        return copy.deepcopy(self)

    
    def crop(self, bins_range):
        """Crop projection"""
        if self.is_compressed(): 
            print "Cropping of PET_Projection currently only works with uncompressed data. Please implement for compressed data. "
            return 
        A = bins_range[0]
        B = bins_range[1]
        data = self.data[:,:,A:B,:]
        self.binning.size_u = (1.0*self.binning.size_u) / self.binning.N_u * (B-A)
        self.binning.N_u = B-A
        return PET_Projection(self.binning, data=data, time_bins=self.time_bins) 
        
    def apply_noise_Poisson(self):
        self.data = poisson(self.data)
        
    # Overload math operators 

    def __add__(self, other): 
        """Overload the '+' (sum) operator. """
        if other is None: #This makes reconstruction code cleaner, when e.g. scatter and randoms are not specified
            return self
        if self._is_same_sparsity_(other):             
            other = other.data 
        out = self.copy()
        out.data = out.data + other
        return out

    def __sub__(self, other): 
        """Overload the '-' (subtraction) operator. """
        if other is None: #This makes reconstruction code cleaner, when e.g. scatter and randoms are not specified
            return self
        if self._is_same_sparsity_(other): 
            other = other.data 
        out = self.copy()
        out.data = out.data - other
        return out

    def __mul__(self, other): 
        """Overload the '*' (multiplication) operator. """
        if other is None: #This makes reconstruction code cleaner, when e.g. scatter and randoms are not specified
            return self
        if self._is_same_sparsity_(other): 
            other = other.data 
        out = self.copy()
        out.data = out.data * other
        return out

    def __div__(self, other): 
        """Overload the '/' (division) operator. """
        if other is None: #This makes reconstruction code cleaner, when e.g. scatter and randoms are not specified
            return self
        if self._is_same_sparsity_(other): 
            other = other.data   
        out = self.copy()
        out.data = out.data / other
        return out

    def __radd_(self, other): 
        return self.__add__(other) 

    def __rmul_(self, other): 
        return self.__add__(other)

    def __rsub__(self, other): 
        """Overload the '-' (subtraction) operator. """
        if other is None: #This makes reconstruction code cleaner, when e.g. scatter and randoms are not specified
            return self
        if self._is_same_sparsity_(other): 
            other = other.data 
        out = self.copy()
        out.data = other-out.data 
        return out

    def __rdiv__(self, other): 
        """Overload the '/' (division) operator. """
        if other is None: #This makes reconstruction code cleaner, when e.g. scatter and randoms are not specified
            return self
        if self._is_same_sparsity_(other): 
            other = other.data 
        out = self.copy()
        out.data = other/out.data 
        return out

    def get_subset(self, subsets_matrix):
        indexes = subsets_matrix.flatten() == 1
        data=self.data.swapaxes(0,1).reshape((self.sparsity.N_axial*self.sparsity.N_azimuthal,self.binning.N_u,self.binning.N_v))[indexes,:,:]
        #sparsity = self.sparsity.get_subset(subsets_matrix)
        #offsets = sparsity.offsets
        #locations = sparsity.locations 
        return PET_Projection(self.binning, data, self.sparsity.offsets, self.sparsity.locations, self.time_bins, subsets_matrix)


    def display_geometry(self):
        return display_PET_Projection_geometry()





# occiput 
# Stefano Pedemonte
# Harvard University, Martinos Center for Biomedical Imaging 
# March 2015, Boston, MA



import ilang 
import ilang.Models 
from ilang.Models import Model 
from ilang.Graphs import ProbabilisticGraphicalModel 

__all__ = ['MR_Static_Gaussian','MR_Dynamic_Gaussian','ProbabilisticGraphicalModel']



class MR_Static_Gaussian(Model): 
    variables = {'x':'continuous','k':'continuous','sigma':'continuous'} 
    dependencies = [['x','k','directed'],['sigma','k','directed']]

    def __init__(self, MR_scan, name=None):
        if name == None:  
            name = self.__class__.__name__
        Model.__init__(self, name) 
        # PET scan
        self.MR_scan = MR_scan    
        # small number
        self.EPS = 1e9

    def set_MR_scan(self, MR_scan): 
        self.MR_scan = MR_scan 

    def init(self): 
        pass 

    def log_conditional_probability_x(self,x_): 
        sigma = self.get_value('sigma')
        return 0 

    def log_conditional_probability_gradient_x(self,x_): 
        sigma = self.get_value('sigma')
        #projection_data = self.PET_scan.project(lambda_, alpha) 
        #Nx = lambda_.shape[0]; Ny = lambda_.shape[1]; Nz = lambda_.shape[2]
        #gradient        = self.PET_scan.backproject(self.PET_scan.get_measurement()[0]/(projection_data[0]+self.EPS), Nx,Ny,Nz, alpha) 
        gradient = 0
        return gradient 

    def log_conditional_probability_sigma(self,sigma): 
        x_ = self.get_value('x') 
        return 0 

    def log_conditional_probability_gradient_sigma(self,sigma): 
        x_ = self.get_value('x')  
        return 0
        
    def sample_conditional_probability_k(self): 
        return 0    



class MR_Dynamic_Gaussian(Model): 
    variables = {'x':'continuous','sigma':'continuous','roi_1':'continuous','roi_1':'continuous','k_1':'continuous','k_2':'continuous'} 
    dependencies = [['x','k_1','directed'],['x','k_2','directed'],['sigma','k_1','directed'],['sigma','k_2','directed'],['roi_1','k_1','directed'],['roi_2','k_2','directed']] 

    def __init__(self, MR_scan, name=None): 
        if name == None:  
            name = self.__class__.__name__
        Model.__init__(self, name) 
        # PET scan object: 
        self.MR_scan = MR_scan 
        self.N_time_bins = len(self.MR_scan.time_bins)
        # Variables and dependencies: 
        self._x      = None 
        self._sigma  = None 
        self._rois   = None 
        self.variables = {'x':'continuous', 'sigma':'continuous'} 
        self.dependencies = []
        for t in range(self.N_time_bins): 
            var_name_counts = 'k_'+str(t)
            var_name_roi    = 'roi_'+str(t)
            self.variables[var_name_counts]='continuous'
            self.variables[var_name_roi]='continuous'
            self.dependencies.append(['x',var_name_counts,'directed'])
            self.dependencies.append(['sigma',var_name_counts,'directed']) 
            self.dependencies.append([var_name_roi,var_name_counts,'directed']) 

    def set_MR_scan(self, MR_scan): 
        self.MR_scan = MR_scan 

    def init(self): 
        pass 

    def log_conditional_probability_lambda(self): 
        return 0 

    def log_conditional_probability_gradient_lambda(self): 
        return numpy.zeros([100,1])

    def log_conditional_probability_alpha(self): 
        return 0 

    def log_conditional_probability_gradient_alpha(self): 
        return numpy.zeros([100,1])
        
    def sample_conditional_probability_z(self): 
        return 0    





        



# occiput 
# Stefano Pedemonte
# Aalto University, School of Science, Helsinki
# Oct 2013, Helsinki 
# Martinos Center for Biomedical Imaging, Harvard University/MGH, Boston
# Dec. 2013, Boston


#__all__ = ['uniform_sphere','uniform_cylinder','uniform_spheres_ring']



from occiput.Core import Image3D as __Image3D
try: 
    from NiftyPy.NiftyRec import ET_spherical_phantom as __ET_spherical_phantom
    from NiftyPy.NiftyRec import ET_cylindrical_phantom as __ET_cylindrical_phantom
    from NiftyPy.NiftyRec import ET_spheres_ring_phantom as __ET_spheres_ring_phantom
except: 
    has_NiftyPy = False
    print "Please install NiftyPy"
else: 
    has_NiftyPy = True

class InstallationError(Exception):
    def __init__(self, message):
        Exception.__init__(self, message)

        

def uniform_sphere(voxels=[256,256,256],size=[1,1,1],center=[0.5,0.5,0.5],radius=0.2,inner_value=1.0,outer_value=0.0): 
    """Create volume (3D numpy array) with uniform value within a spherical region. """
    if not has_NiftyPy: 
        raise InstallationError("Please install NiftyPy to enable function uniform_sphere(..)")
    return __Image3D(__ET_spherical_phantom(voxels,size,center,radius,inner_value,outer_value)) 

def uniform_cylinder(voxels=[256,256,256],size=[1,1,1],center=[0.5,0.5,0.5],radius=0.3,length=0.7,axis=1,inner_value=1.0,outer_value=0.0): 
    """Create volume (3D numpy array) with uniform value within a spherical region. """
    if not has_NiftyPy: 
        raise InstallationError("Please install NiftyPy to enable function uniform_cylinder(..)")
    return __Image3D(__ET_cylindrical_phantom(voxels,size,center,radius,length,axis,inner_value,outer_value)) 
    
def uniform_spheres_ring(voxels=[256,256,256],size=[1,1,1],center=[0.5,0.5,0.5],ring_radius=0.2,min_sphere_radius=0.01,max_sphere_radius=0.1,N_spheres=6,inner_value=1.0,outer_value=0.0,taper=0,axis=0): 
    """Create volume (3D numpy array) with uniform value within a spherical region. """
    if not has_NiftyPy: 
        raise InstallationError("Please install NiftyPy to enable function uniform_spheres_ring(..)")
    return __Image3D(__ET_spheres_ring_phantom(voxels,size,center,ring_radius,min_sphere_radius,max_sphere_radius,N_spheres,inner_value,outer_value,taper,axis))





# occiput 
# Stefano Pedemonte
# Aalto University, School of Science, Helsinki
# Oct 2014, Helsinki 
# Harvard University, Martinos Center for Biomedical Imaging 
# Boston, MA, USA
# April 2014
# Nov. 2015 

# Import an interfile projection as a PET_Projection object of occiput and export. 


__all__ = ["import_interfile_projection", "export_interfile_projection", "import_PET_Projection"] 


from occiput.Reconstruction.PET.PET_projection import PET_Projection, Binning, PET_Projection_Sparsity
from interfile import Interfile
import h5py
import os
from numpy import isscalar, linspace, int32, uint32, ones, zeros, pi, sqrt, float32, float64, where, ndarray, nan
from numpy import inf, asarray, concatenate, fromfile, maximum, exp, asfortranarray, fliplr, transpose  


def import_interfile_projection_data(headerfile='', datafile='',load_time=False):   #FIXME: this should be in the Interfile package
        F = Interfile.load(headerfile)
        
        if F.has_key('matrix size[1]'): 
            N_planes    = F['matrix size[1]']['value']
            N_axial     = F['matrix size[2]']['value']
            N_sinograms = F['matrix size[3]']['value']        
        else:
            N_planes    = F['matrix size [1]']['value']
            N_axial     = F['matrix size [2]']['value']
            N_sinograms = F['matrix size [3]']['value']    
        if datafile == '': 
            datafile1 = headerfile.replace(headerfile.split(os.sep)[-1],F['name of data file']['value'])
            datafile2 = headerfile.replace('.s.hdr','.s') 
            datafile2 = datafile2.replace('.h33','.a')
            datafile3 = headerfile.replace('.h33','.s')
            try: 
                data = fromfile(datafile1,dtype=float32)
            except: 
                try: 
                    data = fromfile(datafile2,dtype=float32)
                except: 
                    try: 
                        data = fromfile(datafile3,dtype=float32)
                    except: 
                        print "Data file not found."
        else: 
            data = fromfile(datafile,dtype=float32)
        data = data.reshape([N_sinograms,N_axial,N_planes])
        if load_time: 
            try: 
                duration = int32([0,F['image duration']['value']])*1000
            except: 
                print "Unable to load image (sinogram) duration. This may determine an incorrect scale and use of randoms and scatter when reconstructing. Set .time_bins manually. "
                duration = int32([0,0])
        else: 
            duration = int32([0,0])
        return data, duration

def import_interfile_projection(headerfile, binning, michelogram, datafile='' ,invert=False, vmin=0.00, vmax=1e10,load_time=False): 
        data, duration = import_interfile_projection_data(headerfile, datafile, load_time=load_time) 
        N_planes    = data.shape[2]
        N_axial     = data.shape[1]
        N_sinograms = data.shape[0]

        N_azim      = michelogram.n_segments
        z_elements = int32([michelogram.segments_sizes[0]])
        for i in range((N_azim-1)/2): 
            z_elements = concatenate([z_elements,int32([michelogram.segments_sizes[i+1],michelogram.segments_sizes[i+1]])])
        
        projection = zeros([binning.N_axial, binning.N_azimuthal, binning.N_u,  binning.N_v], dtype=float32, order="C")
        indexes_azim = [5,6,4,7,3,8,2,9,1,10,0] 
        index0 = 0

        for i in range(N_azim): 
            index1 = index0 + z_elements[i]
            data_azim = data[index0:index1,:,:] 
            for j in range(N_axial): 
                plane = zeros([projection.shape[2],projection.shape[3]],order='F')
                offset = (projection.shape[3]-(index1-index0))/2
                plane[:,offset:offset+index1-index0] = data_azim[:,j,:].squeeze().transpose()
                projection[j,indexes_azim[i],:,:] = fliplr(plane)
            index0 += z_elements[i]
            
        # flip azimuthally - this makes it consistent with Occiput's routines that import from Petlink32 listmode 
        projection2 = zeros(projection.shape,dtype=float32,order="C")
        projection2[:,5,:,:] = projection[:,5,:,:] 
        projection2[:,4,:,:] = projection[:,6,:,:] 
        projection2[:,3,:,:] = projection[:,7,:,:] 
        projection2[:,2,:,:] = projection[:,8,:,:] 
        projection2[:,1,:,:] = projection[:,9,:,:] 
        projection2[:,0,:,:] = projection[:,10,:,:] 
        projection2[:,6,:,:] = projection[:,4,:,:] 
        projection2[:,7,:,:] = projection[:,3,:,:] 
        projection2[:,8,:,:] = projection[:,2,:,:] 
        projection2[:,9,:,:] = projection[:,1,:,:] 
        projection2[:,10,:,:] = projection[:,0,:,:]
        
        sparsity = PET_Projection_Sparsity(binning.N_axial, binning.N_azimuthal, binning.N_u,  binning.N_v)
        ## invert, except where values are 0 
        # if there are broken detectors, disable them (set sensitivity to 0) 
        if invert: 
            projection_fixed = projection2.copy() 
            projection_fixed[projection_fixed<vmin]=0.0 
            projection_fixed[projection_fixed>vmax]=vmax 

            projection_inv = zeros(projection2.shape)
            projection_inv[projection_fixed!=0] = 1.0/projection_fixed[projection_fixed!=0]

            P = PET_Projection( binning, projection_inv, sparsity.offsets, sparsity.locations, duration) 
        else:
            P = PET_Projection( binning, projection2, sparsity.offsets, sparsity.locations, duration) 
        return P 


def export_interfile_projection(sinogram_data_file, projection_data, binning, michelogram, invert=False): 
    # projection_data has size  N_axial, N_azimuthal, N_u, N_v      dtype=float32     order="C"
    # export as  N_sinograms, N_axial, N_planes 
    
    # FIXME: need to flip azimuthally as in import_interfile 

    N_planes = binning.N_u
    N_axial  = binning.N_axial
    N_sinograms = int32(michelogram.segments_sizes[0]) + 2*int32(michelogram.segments_sizes[1::]).sum()
    data = zeros([N_sinograms, N_axial, N_planes], dtype=float32, order='C') 
    
    N_azim      = michelogram.n_segments
    z_elements = int32([michelogram.segments_sizes[0]])
    for i in range((N_azim-1)/2): 
        z_elements = concatenate([z_elements,int32([michelogram.segments_sizes[i+1],michelogram.segments_sizes[i+1]])])
    indexes_azim = [5,6,4,7,3,8,2,9,1,10,0]
    index0=0 

    for i in range(N_azim):
        index1 = index0 + z_elements[i]
        data_azim = zeros( [index1-index0,data.shape[1],data.shape[2]], dtype=float32, order='C' )
        for j in range(N_axial): 
            plane = fliplr( projection_data[j,indexes_azim[i],:,:] )
            offset = (projection_data.shape[3]-(index1-index0))/2
            data_azim[:,j,:] = plane[:,offset:offset+index1-index0].transpose() 
        data[index0:index1,:,:] = data_azim
        index0 += z_elements[i]

    if invert: 
        data2 = zeros([N_sinograms, N_axial, N_planes], dtype=float32, order='C') 
        data2[data!=0] = 1.0 / data[data!=0]
    else: 
        data2 = data
    data2.tofile(sinogram_data_file)



def import_PET_Projection( filename ): 
    h5f = h5py.File(filename,'r') 
    offsets   = asarray( h5f['offsets'],order='F' ) 
    locations = asarray( h5f['locations'],order='F'  )
    try: 
        data      = asarray( h5f['data'],order='F'  )
    except: 
        data      = asarray( h5f['static_data'],order='F'  ) # compatibility with first version - deprecate at some point
    try: 
        time_bins = asarray( h5f['time_bins'],order='F'  )
    except: 
        time_bins = int32([0,0])                   # compatibility with first version - deprecate at some point
    binning = Binning(    {'n_axial':                int32( h5f['n_axial'] ), 
                           'n_azimuthal':            int32( h5f['n_azimuthal'] ), 
                           'angles_axial':           float32( h5f['angles_axial'] ), 
                           'angles_azimuthal':       float32( h5f['angles_azimuthal'] ), 
                           'size_u':                 float32( h5f['size_u'] ), 
                           'size_v':                 float32( h5f['size_v'] ),
                           'n_u':                    int32( h5f['n_u'] ),
                           'n_v':                    int32( h5f['n_v'] ) })
    h5f.close() 
    return PET_Projection(binning, data, offsets, locations, time_bins) 





# occiput 
# Stefano Pedemonte
# Martinos Center for Biomedical Imaging, Harvard University/MGH, Boston
# Sep. 2015, Boston 


__all__ = ["Michelogram"] 


from occiput.Visualization import ipy_table, has_ipy_table, svgwrite, has_svgwrite 
from PIL import ImageDraw
from PIL import Image as PIL 
from numpy import isscalar, linspace, int32, uint32, ones, zeros, pi, sqrt, float32, float64, where, ndarray, nan
from numpy import inf, asarray, concatenate, fromfile, maximum, exp, asfortranarray, fliplr, transpose 


DEFAULT_SPAN = 11 


class Michelogram: 
    def __init__(self, n_rings=64, span=DEFAULT_SPAN, max_ring_difference=60): 
        span = abs(int(span))
        max_ring_difference = abs(int(max_ring_difference))
        n_rings = abs(int(n_rings))
        if (span/2)*2 == span: 
            span = span+1
            print "Span must be odd, using span %d"%span
        if not self._verify_max_ring_difference(span, n_rings, max_ring_difference): 
            max_ring_difference_actual = self._largest_compatible_max_ring_difference(span, n_rings, max_ring_difference)
            compatible_max_ring_diff = self._compatible_max_ring_differences(span, n_rings)
            if compatible_max_ring_diff is []: 
                print "Michelogram: No 'maximum_ring_difference' exists compatible with given 'span' and 'n_rings'. "
            elif max_ring_difference_actual is None: 
                print "Michelogram: 'maximum_ring_difference' %d too small, using %d. "%(max_ring_difference,compatible_max_ring_diff[0])
                max_ring_difference = compatible_max_ring_diff[0]
            else: 
                print "Michelogram: 'maximum_ring_difference' %d is not compatible with 'span' %d. Compatible values are "%(max_ring_difference,span)+str(compatible_max_ring_diff)+". \nUsing %d"%max_ring_difference_actual
                max_ring_difference = max_ring_difference_actual
        self.n_rings = n_rings
        self.span = span 
        self.max_ring_difference = max_ring_difference 
        self.n_segments  = None 
        self.segments_sizes = None 
        self._svg_string = None 
        self._make()
        self._make_svg()

    def _make(self): 
        n_rings = self.n_rings
        n_rings_small = self.span/2
        n_rings_large = n_rings_small+1 
        if self.span>1: 
            n_segments_per_side = (self.max_ring_difference+1-self.span/2)/self.span 
        else: 
            n_segments_per_side = self.max_ring_difference
        n_segments = n_segments_per_side * 2 + 1
        segments_sizes = zeros([n_segments_per_side+1,],dtype=int32)
        if self.span>1: 
            segments_sizes[0] = 2*n_rings-1 
            if n_segments_per_side>0: 
                segments_sizes[1] = segments_sizes[0] - self.span - 1
                for i in range(n_segments_per_side-1): 
                    segments_sizes[2+i]=segments_sizes[1+i]-2*self.span
        else: 
            segments_sizes[0] = n_rings
            for i in range(n_segments_per_side): 
                segments_sizes[i+1]=segments_sizes[i]-1           
        self.n_segments = n_segments
        self.segments_sizes    = segments_sizes

        self.michelogram_sinogram = -1*ones([n_rings,n_rings],order = 'F',dtype=int32)
        self.michelogram_plane    = -1*ones([n_rings,n_rings],order = 'F',dtype=int32)

        if self.span > 1: 
            # -1- right sinograms  
            sinogram_index = int(n_segments)/2
            for X in range(0,self.max_ring_difference,self.span): 
                if sinogram_index == int(n_segments)/2: 
                    plane_index_offset = -n_rings_small
                else: 
                    plane_index_offset = 0
                self._fill_michelogram(X, -n_rings_small, sinogram_index, plane_index_offset)
                sinogram_index += 1

            # -2- left sinograms 
            sinogram_index = int(n_segments)/2-1
            for Y in range(n_rings_large,self.max_ring_difference,self.span): 
                plane_index_offset = 0
                self._fill_michelogram(0, Y, sinogram_index, plane_index_offset)
                sinogram_index -= 1
        else: 
            # -1- right sinograms 
            for seg in range((self.n_segments+1)/2): 
                for y in range(self.segments_sizes[seg]): 
                    x = y+seg
                    self.michelogram_sinogram[x,y] = seg + (self.n_segments+1)/2 - 1
                    self.michelogram_plane[x,y] = y
            # -2- left sinograms 
            for seg in range((self.n_segments+1)/2): 
                for x in range(self.segments_sizes[seg]): 
                    y = x+seg
                    self.michelogram_sinogram[x,y] = (self.n_segments+1)/2 - seg - 1
                    self.michelogram_plane[x,y] = x

    def _verify_max_ring_difference(self, span, n_rings, max_ring_difference): 
        return max_ring_difference in self._compatible_max_ring_differences(span, n_rings)

    def _compatible_max_ring_differences(self, span, n_rings): 
        c = [span/2]
        while True: 
            cnew = c[-1]+span
            if cnew < n_rings:
                c.append(cnew)
            else: 
                break
        return c 

    def _largest_compatible_max_ring_difference(self, span, n_rings, max_ring_difference):
        c = self._compatible_max_ring_differences(span, n_rings)
        for i in range(len(c)):
            if c[len(c)-1-i] < max_ring_difference: 
                return c[len(c)-1-i]
        return None
    
    def _fill_michelogram(self, x_start, y_start, sinogram_index, plane_index_offset): 
        n_rings = self.n_rings
        n_rings_small = self.span/2
        n_rings_large = n_rings_small+1 
        for X in range(n_rings + n_rings_small): 
            # long line
            plane = 2*X + plane_index_offset
            for line_index in range(n_rings_large): 
                x = x_start + X - line_index
                y = y_start + X + line_index
                if x>=0 and x<n_rings and y>=0 and y<n_rings: 
                    self.michelogram_sinogram[x,y] = sinogram_index
                    self.michelogram_plane[x,y] = plane
            # short line
            plane = 2*X + 1 + plane_index_offset
            for line_index in range(n_rings_small): 
                x = x_start + X - line_index
                y = y_start + X + 1 + line_index
                if x>=0 and x<n_rings and y>=0 and y<n_rings: 
                    self.michelogram_sinogram[x,y] = sinogram_index
                    self.michelogram_plane[x,y] = plane 

    def display_michelogram(self, cmap='gray', figsize=[16,8]): 
        import matplotlib.pyplot as plt
        plt.figure(figsize=figsize)
        plt.subplot(1,2,1); 
        plt.imshow(self.michelogram_plane,interpolation='nearest')
        plt.set_cmap(cmap)
        plt.subplot(1,2,2); 
        plt.imshow(self.michelogram_sinogram,interpolation='nearest')
        plt.set_cmap(cmap)

    def _make_svg(self): 
        if not has_svgwrite: 
            self._svg_string = None 
            return self._svg_string 
        w = '100%'
        h = '100%'
        dwg = svgwrite.Drawing('michelogram.svg',size=(w,h), profile='full', debug=True)
        dwg.viewbox(width=100, height=100)

        # blank table 
        rect = dwg.add(dwg.rect(insert=(0, 0), size=(100, 100), rx=0.0, ry=0.0))
        rect.fill('grey',opacity=0.1).stroke('black',width=0.1,opacity=0.001)
        #for ... 
        #line = dwg.add(dwg.polyline([(10, 10), (10, 100), (100, 100), (100, 10), (10, 10)],stroke='black', fill='none'))

        # fill table with Michelogram data 

        self._svg_string = dwg.tostring() 
        return self._svg_string

    def _repr_svg_(self): 
        return self._svg_string 









# occiput 
# Stefano Pedemonte 
# April 2014 
# Harvard University, Martinos Center for Biomedical Imaging 
# Boston, MA, USA 
# Nov. 2015 

__all__ = ['vNAV_MPRage','load_vnav_mprage'] 

from PIL import Image 
from occiput.Core import Image3D, Transform_Affine 
import numpy 
import nibabel 
import dicom 
import matplotlib.pyplot as plt
import os 
import copy 
from occiput.Core import transformations as tr


BOX_MIN      = [-50.0,-50.0,-50.0]
BOX_MAX      = [50.0,50.0,50.0]
THRESHOLD_MM = 1.1
LINE_COLOR   = "#2f8dff"

def quaternion_to_rotation(q,axes='sxyz'): 
    return tr.euler_from_quaternion(q,axes) 

def angle_axis_to_quaternion(angle_rad,axis):
    f = numpy.sin(0.5*angle_rad)
    quaternion = numpy.asarray( [numpy.cos(0.5*angle_rad),f*axis[0],f*axis[1],f*axis[2]] )
    return quaternion

def angle_axis_to_rotation(angle_rad,axis): 
    quaternion = angle_axis_to_quaternion(angle_rad,axis)
    rotation = quaternion_to_rotation(quaternion) 
    return rotation

def affine_from_quaternion(q,axes='sxyz'): 
    pass 

def quaternion_from_matrix(affine):
    return tx.quaternion_from_matrix(affine)

def rad_to_deg(rad): 
    return numpy.asarray(rad)*180.0/numpy.pi
    
def deg_to_rad(deg): 
    return numpy.asarray(deg)/180.0*numpy.pi




class vNAV_MPRage(): 
    def __init__(self,path=None, from_dicom_comments=True, files_start_with=None, files_end_with='.dcm'): 
        self._n_time_points     = 0 
        self._duration          = [] 
        self._motion            = [] 
        if path  is not None: 
            self.load_data_files(path, from_dicom_comments, files_start_with, files_end_with) 

    def get_volume_dicom(self, index): 
        dcm_file = self._paths[index] 
        f = dicom.read_file(dcm_file) 
        return f 

    def get_motion_quaternion(self, index): 
        motion_affine = self.get_motion_affine(index) 
        return quaternion_from_matrix(motion_affine)

    def get_motion_affine(self, index): 
        return self._motion[index] 

    def get_n_time_points(self): 
        return self._n_time_points 

    def get_duration(self,index): 
        return self.duration[index] 

    def load_data_files(self, path, from_dicom_comments=True, files_start_with=None, files_end_with=None, exclude_files_end_with=['.dat','.txt','.py','.pyc','.nii','.gz','.png','.jpg','.jpeg','.eps','.hdr','.l'] ):
        """Load vNAV dicom files from given path and extract motion information. 
        If from_dicom_comments==True, use the information stored in the dicom comments. 
        If from_dicom_comments==False, use the information stored in the dicom MOCO field. 
        As of April 2014, these two express motion in different coordinate systems. 
        The dicom comments report absolute motion in scanner coordinates (origin is the magnetic iso-center of the scanner).
        The dicom MOCO field is currently for proprietary use (??). """ 

        self._n_time_points     = 0 
        self._duration          = [] 
        self._motion            = [] 

        self._paths = []
        self._tx = []; self._ty = []; self._tz = []; 
        self._rx = []; self._ry = []; self._rz = []; 
        self._tx_comm = []; self._ty_comm = []; self._tz_comm = [];
        self._rx_comm = []; self._ry_comm = []; self._rz_comm = [];
        self._q0_comm = []; self._q1_comm = []; self._q2_comm = []; self._q3_comm = []; 
        self._a0_comm = []; self._a1_comm = []; self._a2_comm = []; self._a3_comm = []; 
        N=0 

        # pick the first dicom file found in path 
        files = os.listdir(path)
        files.sort() 

        # CASE 1: there exist files named with .dcm extension
        for file_name in files: 
            file_valid = True
            if files_start_with is not None: 
                if not file_name.startswith(files_start_with): 
                    file_valid = False
            if files_end_with is not None: 
                if not file_name.endswith(files_end_with): 
                    file_valid = False 
            for s in exclude_files_end_with: 
                if file_name.endswith(s): 
                    file_valid = False           
            if file_valid: 
                full_path = path+os.sep+file_name
                # read moco information from files 
                self._paths.append(full_path)
                try: 
                    f = dicom.read_file(full_path) 
                except: 
                    print "Could not read file ",full_path
                    return 
                t = f.get(0x00191025).value 
                r = f.get(0x00191026).value 
                self._tx.append(t[0]); self._ty.append(t[1]); self._tz.append(t[2]); 
                self._rx.append(r[0]); self._ry.append(r[1]); self._rz.append(r[2]); 
                motion_dicom_moco = []
                
                # extract moco information stored in the dicom comment field
                if from_dicom_comments: 
                    s = f.get(0x00204000).value 
                    if N: 
                        a = numpy.float32(s.split(' ')[1:5])
                        t = numpy.float32(s.split(' ')[6:9])
                        freq = numpy.float32(s.split(' ')[10])
                        r = angle_axis_to_rotation(a[0],a[1:4]) 
                    else: 
                        t = numpy.float32([0,0,0])
                        r = numpy.float32([0,0,0]) 
                        a = numpy.float32([0,1,0,0])  #FIXME: is this right?
                
                    q = angle_axis_to_quaternion(a.copy()[0],a.copy()[1:4]) 
                    self._a0_comm.append(a[0]); self._a1_comm.append(a[1]); self._a2_comm.append(a[2]); self._a3_comm.append(a[3]); 
                    self._tx_comm.append(t[0]); self._ty_comm.append(t[1]); self._tz_comm.append(t[2])
                    self._q0_comm.append(q[0]); self._q1_comm.append(q[1]); self._q2_comm.append(q[2]); self._q3_comm.append(q[3]); 
                    self._rx_comm.append(r[0]); self._ry_comm.append(r[1]); self._rz_comm.append(r[2]); 


                    tra_mat = tr.translation_matrix(t) 
                    rot_mat = tr.quaternion_matrix(q)
                    motion_dicom_comments = numpy.dot(tra_mat,rot_mat) 

                #xaxis, yaxis, zaxis = [1, 0, 0], [0, 1, 0], [0, 0, 1]
                #Rx = tr.rotation_matrix(r[0], xaxis)
                #Ry = tr.rotation_matrix(r[1], yaxis)
                #Rz = tr.rotation_matrix(r[2], zaxis)
                #rot_mat = tr.concatenate_matrices(Rx, Ry, Rz)
                #rot_mat = Ry.copy()
                #motion_dicom_comments = numpy.dot(tra_mat,rot_mat) 
                #motion_dicom_comments = rot_mat.copy()
                
                N += 1 
                if from_dicom_comments: 
                    self._motion.append(motion_dicom_comments) 
                else: 
                    self._motion.append(motion_dicom_moco) 
                acquisition_number = f.get(0x00200012).value 
                creation_time      = f.get(0x00080013).value
#                print "Acquisition number: ", acquisition_number
#                print "Creation time:      ",creation_time
        self._n_time_points = N

    def _draw_rectangle(self, axis, x, y, alpha=0.2, ec="gray", fc="gray"): #"CornflowerBlue"
        axis.add_patch( plt.Rectangle( (x[0],y[0]) , x[1]-x[0],y[1]-y[0], alpha=alpha, ec=ec, fc=fc, visible=True) )
        #plt.draw() 

    def _draw_rectangles(self, axis, windows, range_y): 
        for ii in range(len(windows)):
            tt = windows[ii]
            yy = (range_y[0],range_y[1])
            self._draw_rectangle(axis,tt,yy)

    def _draw_line(self, axis, x, range_y, color="#ff8d8d", linestyle="dashed", label=""): 
        axis.vlines(x, range_y[0], range_y[1], colors=color, linestyles=linestyle, label=label, visible=True)

    def _draw_events(self, axis, time, range_y, events, color="#ff8d8d", linestyle="dashed", label=""): 
        for t_index in time: 
            if t_index: 
                if events[t_index-1]: 
                    #print "Drawing line: ", t_index, range_y, color, linestyle
                    self._draw_line(axis, t_index, range_y, color, linestyle, label)

    def plot_motion(self, save_to_file=None, display_dicom_comments=True, display_dicom_moco=False, range_mm=(-8,8), range_deg=(-7,7), extract_events_threshold=THRESHOLD_MM, method='box', box_min=BOX_MIN, box_max=BOX_MAX, min_duration=1, line_color=LINE_COLOR, figsize=[10,5]):
        t = range(len(self._tx))

        # make windows:
        if extract_events_threshold is not None: 
            windows = []
            events  = self.extract_motion_events(method, extract_events_threshold, box_min, box_max)
        t_index_start = 0
        for t_index in t:  
            if t_index:     #this excludes the possibility of a motion event at time 0 
                if events[t_index-1]: 
                    t_index_end = t_index-1
                    if t_index_end-t_index_start > min_duration: 
                        windows.append( (t_index_start, t_index_end) )  #end window with frame before a motion event 
                    t_index_start = t_index+1                 #start window with frame after a motion event 
        windows.append( (t_index_start, t[-1]) ) 

        if display_dicom_moco:      
            fig1 = plt.figure(1,figsize=figsize)

            ax1 = fig1.add_subplot(311)
            ax1.plot(t, self._tx, line_color)
            ax1.grid(True)
            ax1.set_ylim( range_mm )
            ax1.set_ylabel('TX [mm]')
            #for label in ax1.get_xticklabels():
            #    label.set_color('r')
            if extract_events_threshold  is not None:
                self._draw_rectangles(ax1,windows,range_mm) 
                self._draw_events(ax1, t, range_mm, events) 

            ax1 = fig1.add_subplot(312)
            ax1.plot(t, self._ty, line_color)
            ax1.grid(True)
            ax1.set_ylim( range_mm )
            ax1.set_ylabel('TY [mm]')
            #for label in ax1.get_xticklabels():
            #    label.set_color('r')
            if extract_events_threshold  is not None:
                self._draw_rectangles(ax1,windows,range_mm) 
                self._draw_events(ax1, t, range_mm, events) 
    
            ax1 = fig1.add_subplot(313)
            ax1.plot(t, self._tz, line_color)
            ax1.grid(True)
            ax1.set_ylim( range_mm )
            ax1.set_ylabel('TZ [mm]')
            #for label in ax1.get_xticklabels():
            #    label.set_color('r')
            if extract_events_threshold  is not None:
                self._draw_rectangles(ax1,windows,range_mm) 
                self._draw_events(ax1, t, range_mm, events) 

#            if save_to_file is not None: 
#                plt.savefig(save_to_file)

            fig2 = plt.figure(2,figsize=figsize)
    
            ax1 = fig2.add_subplot(311)
            ax1.plot(t, rad_to_deg(self._rx), line_color)
            ax1.grid(True)
            ax1.set_ylim( range_deg )
            ax1.set_ylabel('RX [deg]')
            #for label in ax1.get_xticklabels():
            #    label.set_color('r')
            if extract_events_threshold  is not None:
                self._draw_rectangles(ax1,windows,range_deg) 
                self._draw_events(ax1, t, range_deg, events) 
            
            ax1 = fig2.add_subplot(312)
            ax1.plot(t, rad_to_deg(self._ry), line_color)
            ax1.grid(True)
            ax1.set_ylim( range_deg )
            ax1.set_ylabel('RY [deg]')
            #for label in ax1.get_xticklabels():
            #    label.set_color('r')
            if extract_events_threshold  is not None:
                self._draw_rectangles(ax1,windows,range_deg) 
                self._draw_events(ax1, t, range_deg, events) 
                                
            ax1 = fig2.add_subplot(313)
            ax1.plot(t, rad_to_deg(self._rz), line_color)
            ax1.grid(True)
            ax1.set_ylim( range_deg )
            ax1.set_ylabel('RZ [deg]')
            #for label in ax1.get_xticklabels():
            #    label.set_color('r')
            if extract_events_threshold  is not None:
                self._draw_rectangles(ax1,windows,range_deg) 
                self._draw_events(ax1, t, range_deg, events) 
                                
#            if save_to_file is not None: 
#                plt.savefig(save_to_file)
        else: 
            fig1=None
            fig2=None

        if display_dicom_comments: 
            fig3 = plt.figure(3,figsize=figsize)

            #mr = numpy.min([self._rx_comm,self._ry_comm,self._rz_comm])
            #Mr = numpy.max([self._rx_comm,self._ry_comm,self._rz_comm])
            #mt = numpy.min([self._tx_comm,self._ty_comm,self._tz_comm])
            #Mt = numpy.max([self._tx_comm,self._ty_comm,self._tz_comm])
            mr = range_deg[0]
            Mr = range_deg[1]
            mt = range_mm[0]
            Mt = range_mm[1]

            ax1 = fig3.add_subplot(311)
            ax1.set_title("Rotation vs. vNAV frame number")
            ax1.plot(t, rad_to_deg(self._rx_comm), line_color)
            ax1.grid(True)
            ax1.set_ylim( (mr,Mr) )
            ax1.set_ylabel('RX [deg]')
            #for label in ax1.get_xticklabels():
            #    label.set_color('r')
            #print windows
            if extract_events_threshold  is not None:
                self._draw_rectangles(ax1,windows,range_deg) 
                self._draw_events(ax1, t, range_deg, events) 
    
            ax1 = fig3.add_subplot(312)
#            ax1.set_title("Rotation comments Y")
            ax1.plot(t, rad_to_deg(self._ry_comm), line_color)
            ax1.grid(True)
            ax1.set_ylim( (mr,Mr) )
            ax1.set_ylabel('RY [deg]')
            #for label in ax1.get_xticklabels():
            #    label.set_color('r')
            if extract_events_threshold  is not None:
                self._draw_rectangles(ax1,windows,range_deg) 
                self._draw_events(ax1, t, range_deg, events) 
                
            ax1 = fig3.add_subplot(313)
#            ax1.set_title("Rotation comments Z")
            ax1.plot(t, rad_to_deg(self._rz_comm), line_color)
            ax1.grid(True)
            ax1.set_ylim( (mr,Mr) )
            ax1.set_ylabel('RZ [deg]')
            #for label in ax1.get_xticklabels():
            #    label.set_color('r')
            if extract_events_threshold  is not None:
                self._draw_rectangles(ax1,windows,range_deg) 
                self._draw_events(ax1, t, range_deg, events) 
        
#            if save_to_file is not None: 
#                plt.savefig(save_to_file)

            fig4 = plt.figure(4,figsize=figsize)

            ax1 = fig4.add_subplot(311)
            ax1.set_title("Translation vs. vNAV frame number")
            ax1.plot(t, self._tx_comm, line_color)
            ax1.grid(True)
            ax1.set_ylim( (mt,Mt) )
            ax1.set_ylabel('TX [mm]')
            #for label in ax1.get_xticklabels():
            #    label.set_color('r')
            if extract_events_threshold  is not None:
                self._draw_rectangles(ax1,windows,range_mm) 
                self._draw_events(ax1, t, range_mm, events) 
                
            ax1 = fig4.add_subplot(312)
#            ax1.set_title("Translation comments Y")
            ax1.plot(t, self._ty_comm, line_color)
            ax1.grid(True)
            ax1.set_ylim( (mt,Mt) )
            ax1.set_ylabel('TY [mm]')
            #for label in ax1.get_xticklabels():
            #    label.set_color('r')
            if extract_events_threshold  is not None:
                self._draw_rectangles(ax1,windows,range_mm) 
                self._draw_events(ax1, t, range_mm, events) 
                
            ax1 = fig4.add_subplot(313)
#            ax1.set_title("Translation comments Z")
            ax1.plot(t, self._tz_comm, line_color)
            ax1.grid(True)
            ax1.set_ylim( (mt,Mt) )
            ax1.set_ylabel('TZ [mm]')
            #for label in ax1.get_xticklabels():
            #    label.set_color('r')
            if extract_events_threshold  is not None:
                self._draw_rectangles(ax1,windows,range_mm) 
                self._draw_events(ax1, t, range_mm, events) 
                
#            if save_to_file is not None: 
#                plt.savefig(save_to_file)
        else: 
            fig3=None
            fig4=None
            
        plt.show()
        return fig1,fig2,fig3,fig4


    def get_mean_displacement(self, index, method='box', box_min=BOX_MIN, box_max=BOX_MAX): 
        mat = self.get_motion_affine(index)
        mat = Transform_Affine(mat) 
        if method=='box': 
            b = box_min
            B = box_max
            corners = numpy.asarray([ [b[0],b[1],b[2],1], [B[0],b[1],b[2],1], [b[0],B[1],b[2],1], [B[0],B[1],b[2],1], [b[0],b[1],B[2],1], [B[0],b[1],B[2],1], [b[0],B[1],B[2],1], [B[0],B[1],B[2],1] ]).transpose()
            corners_t = mat.left_multiply(corners)
            corners_t[3,:]=0
            corners[3,:]  =0
            #print "CORNERS: ", corners_t
            #dist = numpy.sqrt(((corners-corners_t)**2).sum(0))
            dist = (corners-corners_t).sum(0)
            mean_displ = numpy.mean(dist)
        else: 
            raise "Method to compute mean displacement is unknown. "
        return mean_displ
 
    def get_mean_displacement_variation(self, index, method='box', box_min=BOX_MIN, box_max=BOX_MAX):
        mat  = self.get_motion_affine(index)
        if index > 0: 
            mat0 = self.get_motion_affine(index-1)
        else: 
            mat0 = tr.identity_matrix()
        mat = Transform_Affine(mat) 
        mat0 = Transform_Affine(mat0)
        if method=='box': 
            b = box_min
            B = box_max
            corners = numpy.asarray([ [b[0],b[1],b[2],1], [B[0],b[1],b[2],1], [b[0],B[1],b[2],1], [B[0],B[1],b[2],1], [b[0],b[1],B[2],1], [B[0],b[1],B[2],1], [b[0],B[1],B[2],1], [B[0],B[1],B[2],1] ]).transpose()
            corners_t = mat.left_multiply(corners)
            corners_t[3,:]=0
            corners_t0 = mat0.left_multiply(corners)
            corners_t0[3,:]=0
            dist = numpy.sqrt(((corners_t-corners_t0)**2).sum(0))
            #dist = (corners-corners_t).sum(0)
            mean_displ = numpy.mean(dist)
        else: 
            raise "Method to compute mean displacement is unknown. "
        return mean_displ

    def get_mean_displacement_variation_since_time(self, index_new, index_old, method='box', box_min=BOX_MIN, box_max=BOX_MAX):
        mat  = self.get_motion_affine(index_new)
        mat0 = self.get_motion_affine(index_old)
        mat = Transform_Affine(mat) 
        mat0 = Transform_Affine(mat0)
        if method=='box': 
            b = box_min
            B = box_max
            corners = numpy.asarray([ [b[0],b[1],b[2],1], [B[0],b[1],b[2],1], [b[0],B[1],b[2],1], [B[0],B[1],b[2],1], [b[0],b[1],B[2],1], [B[0],b[1],B[2],1], [b[0],B[1],B[2],1], [B[0],B[1],B[2],1] ]).transpose()
            corners_t = mat.left_multiply(corners)
            corners_t[3,:]=0
            corners_t0 = mat0.left_multiply(corners)
            corners_t0[3,:]=0
            dist = numpy.sqrt(((corners_t-corners_t0)**2).sum(0))
            #dist = (corners-corners_t).sum(0)
            mean_displ = numpy.mean(dist)
        else: 
            raise "Method to compute mean displacement is unknown. "
        return mean_displ

        
    def extract_motion_events(self, method='box', threshold=THRESHOLD_MM, box_min=BOX_MIN, box_max=BOX_MAX):
        t = range(self.get_n_time_points())
        is_event   = numpy.zeros(len(t)-1)
        t_index_old = 0
        for t_index in t[1:]: 
#             mean_displ = self.get_mean_displacement_variation(t_index, method, box_min, box_max ) 
#             if numpy.sqrt((mean_displ)**2) >= threshold: 

             mean_displ = self.get_mean_displacement_variation_since_time(t_index, t_index_old, method, box_min, box_max )
             if numpy.sqrt((mean_displ)**2) >= threshold: 
                 t_index_old = numpy.copy(t_index)
                 is_event[t_index-1] = 1
             else: 
                 is_event[t_index-1] = 0 
        return is_event 


    def plot_mean_displacement(self, method='box', box_min=BOX_MIN,box_max=BOX_MAX, save_to_file=None, plot_zero=False, extract_events_threshold=THRESHOLD_MM, plot_range=[None,None], line_color=LINE_COLOR, min_duration=1, figsize=[10,5] ): 
        t = range(self.get_n_time_points())
        mean_displ = numpy.zeros(len(t))
        mean_displ_var = numpy.zeros(len(t))
        mean_displ_var_since_event = numpy.zeros(len(t))
        
        if extract_events_threshold is not None: 
            events = self.extract_motion_events(method, extract_events_threshold, box_min, box_max)

        t_index_old = 0
        for t_index in t: 
             mean_displ[t_index] = self.get_mean_displacement(t_index, method, box_min, box_max ) 
             mean_displ_var[t_index] = self.get_mean_displacement_variation(t_index, method, box_min, box_max ) 
             mean_displ_var_since_event[t_index] = self.get_mean_displacement_variation_since_time(t_index, t_index_old, method, box_min, box_max ) 
             if t_index: 
                 if events[t_index-1] == 1: 
                     t_index_old = t_index-1
        if not plot_zero: 
            t = t[1:]
            mean_displ = mean_displ[1:]
            mean_displ_var = mean_displ_var[1:]
            mean_displ_var_since_event = mean_displ_var_since_event[1:]

        # make windows:
        if extract_events_threshold is not None: 
            windows = []
            events  = self.extract_motion_events(method, extract_events_threshold, box_min, box_max)
        t_index_start = 0
        for t_index in t:  
            if t_index:     #this excludes the possibility of a motion event at time 0 
                if events[t_index-1]: 
                    t_index_end = t_index-1
                    if t_index_end-t_index_start > min_duration: 
                        windows.append( (t_index_start, t_index_end) )  #end window with frame before a motion event 
                    t_index_start = t_index+1                 #start window with frame after a motion event 
        windows.append( (t_index_start, t[-1]) ) 
        
#        mean_displ[numpy.where(mean_displ==0)]=-1000
#        mean_displ_var[numpy.where(mean_displ_var==0)]=-1000
#        mean_displ_var_since_event[numpy.where(mean_displ_var_since_event==0)]=-1000
        
        fig = plt.figure(5,figsize=figsize)

        ax1 = fig.add_subplot(311)
        ax1.set_title("Times frames - vNAV")
        ax1.plot(t,mean_displ, line_color)
        ax1.grid(True)
        if plot_range[0] is None: 
            pr0=copy.copy(mean_displ.min())
        else: 
            pr0=copy.copy(plot_range[0])
        if plot_range[1] is None: 
            pr1=copy.copy(mean_displ.max())
        else: 
            pr1=copy.copy(plot_range[1])
        ax1.set_ylim( [pr0,pr1] )
        ax1.set_ylabel('disp [mm]')
        if extract_events_threshold is not None: 
            ax1.hold(1)
            E = events*mean_displ
            E[numpy.where(E<0.1)]=-1000
            ax1.plot(t,E,'r.')
            #ax1.plot(t,0.5*(events*(mean_displ.max()-mean_displ.min())+mean_displ.min()),'r.')
        #for label in ax1.get_xticklabels():
        #    label.set_color('r')
        self._draw_rectangles(ax1,windows,[pr0,pr1]) 
        self._draw_events(ax1, t, [pr0,pr1], events) 


        ax1 = fig.add_subplot(312)
        #ax1.set_title("Mean displacement delta ")
        ax1.plot(t,mean_displ_var, line_color)
        ax1.grid(True)
        if plot_range[0] is None: 
            pr0=copy.copy(mean_displ_var.min())
        else: 
            pr0=copy.copy(plot_range[0])
        if plot_range[1] is None: 
            pr1=copy.copy(mean_displ_var.max())
        else: 
            pr1=copy.copy(plot_range[1])
        ax1.set_ylim( [pr0,pr1] )
        ax1.set_ylabel('delta [mm]')
        if extract_events_threshold is not None: 
            ax1.hold(1)
            E = events*mean_displ_var
            E[numpy.where(E<0.1)]=-1000
            ax1.plot(t,E,'r.')
            #ax1.plot(t,0.5*(events*(mean_displ.max()-mean_displ.min())+mean_displ.min()),'r.')
        #for label in ax1.get_xticklabels():
        #    label.set_color('r')
        self._draw_rectangles(ax1,windows,[pr0,pr1]) 
        self._draw_events(ax1, t, [pr0,pr1], events)             

        ax1 = fig.add_subplot(313)
        #ax1.set_title("Mean displacement event")
        ax1.plot(t,mean_displ_var_since_event, line_color)
        ax1.grid(True)
        if plot_range[0] is None: 
            pr0=copy.copy(mean_displ_var_since_event.min())
        else: 
            pr0=copy.copy(plot_range[0])
        if plot_range[1] is None: 
            pr1=copy.copy(mean_displ_var_since_event.max())
        else: 
            pr1=copy.copy(plot_range[1])
        ax1.set_ylim( [pr0,pr1] )
        ax1.set_ylabel('event [mm]')
        if extract_events_threshold is not None: 
            ax1.hold(1)
            E = events*mean_displ_var_since_event
            E[numpy.where(E<0.1)]=-1000
            ax1.plot(t,E,'r.')
            #ax1.plot(t,0.5*(events*(mean_displ.max()-mean_displ.min())+mean_displ.min()),'r.')
        #for label in ax1.get_xticklabels():
        #    label.set_color('r')
        self._draw_rectangles(ax1,windows,[pr0,pr1]) 
        self._draw_events(ax1, t, [pr0,pr1], events) 

        if save_to_file is not None: 
            plt.savefig(save_to_file)
        plt.show() 
        return fig

    def plot_quaternion(self, save_to_file=None, plot_range=[None,None], line_color=LINE_COLOR, figsize=[10,5]): 
        t = range(self.get_n_time_points())[1:]
        s = rad_to_deg(numpy.asarray(self._a0_comm))[1:]
        
        fig = plt.figure(6,figsize=figsize)
        ax1 = fig.add_subplot(211)
        ax1.set_title("Rotation agnle [deg] vs. vNAV frame number")
        ax1.plot(t,s, line_color)
        ax1.grid(True)
        if plot_range[0] is None: 
            plot_range[0]=s.min()
        if plot_range[1] is None: 
            plot_range[1]=s.max()
        ax1.set_ylim( plot_range )
        ax1.set_ylabel('Rotation angle [deg]')
        #for label in ax1.get_xticklabels():
        #    label.set_color('r')

        arc = numpy.zeros(self.get_n_time_points())
        v0  = numpy.asarray( self._a1_comm[0],self._a2_comm[0],self._a3_comm[0] )
        for t in range(self.get_n_time_points()): 
            vt  = numpy.asarray( self._a1_comm[t],self._a2_comm[t],self._a3_comm[t] )
            arc[t] = numpy.dot( numpy.transpose(v0), vt )
        ax1 = fig.add_subplot(212)
        ax1.set_title("Arc vs. vNAV frame number")
        t = range(self.get_n_time_points())
        ax1.plot(t,arc, line_color)
        ax1.grid(True)
        plot_range[0]=arc.min()-0.0001
        plot_range[1]=arc.max()+0.0001
        ax1.set_ylim( plot_range )
        ax1.set_ylabel('Arc [steradians]')
        #for label in ax1.get_xticklabels():
        #    label.set_color('r')
    
        if save_to_file is not None: 
            plt.savefig(save_to_file)
        plt.show() 
        return fig


    def _repr_html_(self): 
        self.plot_mean_displacement() 

    
def load_vnav_mprage(path, from_dicom_comments=True, files_start_with=None, files_end_with=None): 
    return vNAV_MPRage(path, from_dicom_comments, files_start_with, files_end_with)




# occiput  
# Stefano Pedemonte
# Harvard University, Martinos Center for Biomedical Imaging 
# April. 2015, Boston, MA

from math import pi 


def millisec_to_min_sec(ms):
    """Convert millisecond [integer] to string: min sec msec."""
    totsec = int(ms)/1000
    sec = totsec%60
    min = int(ms)/1000/60 
    msec = ms%(1000)
    return str(min).zfill(2)+" [min]  "+ str(sec).zfill(2)+" [sec]  " + str(msec).zfill(3)+" [msec]"

def pretty_print_large_number(number):
    """Given a large number, it returns a string of the sort:  '10.5 Thousand' or '12.3 Billion'. """
    s = str(number).ljust(12)
    if number > 0 and number < 1e3: 
        pass 
    elif number >= 1e3 and number < 1e6: 
        s = s + "  (%3.1f Thousand)"%(number*1.0/1e3)
    elif number >= 1e6 and number < 1e9: 
        s = s + "  (%3.1f Million)"%(number*1.0/1e6)
    elif number >= 1e9 and number < 1e12: 
        s = s + "  (%3.1f Billion)"%(number*1.0/1e9)
    elif number >= 1e12 and number < 1e15: 
        s = s + "  (%3.1f Trillion)"%(number*1.0/1e12)
    return s

def print_percentage(number): 
    """Given a number between 0 and 1, it returns a string indicating percentage: e.g. 10.2 %%"""
    return "%2.2f %%"%((1.0*number)*100) 

def rad_to_deg(rad): 
    """Convert radians to degrees."""
    return rad * 180.0 / pi

def deg_to_rad(deg): 
    """Convert degrees to radians."""
    return deg * pi / 180.0 

def array_to_string(array, format="%3.3f "): 
    """Return a string from an array, with given formatting. """
    s = ""
    for i in array: 
        s = s+format%i 
    return s

    




def is_in_ipynb(): 
    try:
        from IPython import get_ipython
        cfg = get_ipython().config 
        if cfg['IPKernelApp']['parent_appname'] == 'ipython-notebook':
            return True
        else:
            return False
    except:
        return False




# occiput 
# Stefano Pedemonte 
# Oct 2015
# Harvard University, Martinos Center for Biomedical Imaging 
# Boston, MA, USA 


from PIL import Image
import numpy
import uuid

from occiput.global_settings import is_gpu_enabled
from DisplayNode import DisplayNode
import Colors as C
from IPython.display import HTML, display, Javascript
from ipynb import is_in_ipynb



class InstallationError(Exception):
    def __init__(self, message):
        Exception.__init__(self, message)

        
    
    
    
class ProgressBar(): 
    def __init__(self, height='6px', width='100%%', background_color=C.LIGHT_BLUE, foreground_color=C.BLUE): 
        self._percentage = 0.0
        self._divid = str(uuid.uuid4()) 
        self.visible = False
        if is_in_ipynb(): 
            self.set_display_mode("ipynb")
        else: 
            self.set_display_mode("text")
        self._pb = HTML(
            """
            <div style="border: 1px solid white; width:%s; height:%s; background-color:%s">
                <div id="%s" style="background-color:%s; width:0%%; height:%s"> </div>
            </div> 
            """ % ( width, height, background_color, self._divid, foreground_color, height))

    def show(self): 
        if self.mode == "ipynb": 
            display(self._pb)
        self.visible = True 
        
    def set_display_mode(self, mode="ipynb"): 
        self.mode=mode 

    def set_percentage(self,percentage):
        if not self.visible: 
            self.show()
        if percentage < 0.0:
            percentage = 0.0
        if percentage > 100.0:
            percentage = 100.0 
        percentage = int(percentage)
        self._percentage = percentage
        if self.mode == "ipynb": 
            display(Javascript("$('div#%s').width('%i%%')" % (self._divid, percentage)))
        else: 
            print "%2.1f / 100"%percentage

    def get_percentage(self):
        return self._percentage







class MultipleVolumes(): 
    def __init__(self,volumes,axis=0,shrink=256,rotate=90,subsample_slices=None,scales=None,open_browser=None): 
        self.volumes = volumes 
        self._axis = axis
        self._shrink = shrink
        self._rotate = rotate
        self._subsample_slices = subsample_slices
        self._scales = scales
        self._open_browser = open_browser
        self._progress_bar = ProgressBar(height='6px', width='100%%', background_color=C.LIGHT_GRAY, foreground_color=C.GRAY)

    def get_data(self,volume_index): 
        volume = self.volumes[volume_index] 
        if isinstance(volume,numpy.ndarray): 
            return volume
        else:
            # Volume type
            return volume.data
        
    def get_shape(self,volume_index): 
        return self.volumes[volume_index].shape
        
    def export_image(self,volume_index,slice_index,axis=0,normalise=True,scale=None,shrink=None,rotate=0,global_scale=True): 
        #FIXME: handle 4D, 5D, ..
        
        M = self.get_data(volume_index).max()
        m = self.get_data(volume_index).min()
        
        if axis == 0:
            a = numpy.float32(self.get_data(volume_index)[slice_index,:,:].reshape((self.get_shape(volume_index)[1],self.get_shape(volume_index)[2]))) 
        elif axis == 1: 
            a = numpy.float32(self.get_data(volume_index)[:,slice_index,:].reshape((self.get_shape(volume_index)[0],self.get_shape(volume_index)[2])))
        else: 
            a = numpy.float32(self.get_data(volume_index)[:,:,slice_index].reshape((self.get_shape(volume_index)[0],self.get_shape(volume_index)[1])))

        if m>=0: 
            if normalise: 
                if not global_scale: 
                    if scale is None:
                        a = a * 255/(a.max()+1e-9) 
                    else: 
                        a = a * scale *255/(a.max()+1e-9) 
                else: 
                    if scale is None:
                        a = a * 255/(M+1e-9) 
                    else: 
                        a = a * scale *255/(M+1e-9) 
            im = Image.fromarray(a).convert("RGB") 
        else:
            if normalise: 
                if not global_scale:
                    if scale is None:
                        a = a * 512/(a.max()-a.min()+1e-9) 
                    else: 
                        a = a * scale * 512/(a.max()-a.min()+1e-9) 
                else: 
                    if scale is None:
                        a = a * 512/(M-m+1e-9) 
                    else: 
                        a = a * scale * 512/(M-m+1e-9)                     
            blue = a
            red  = a.copy()
            red[numpy.where(red>=0)]  = 0 
            red  = - red
            blue[numpy.where(blue<0)] = 0 
            green = numpy.zeros(red.shape)
            rgb = numpy.zeros((red.shape[0],red.shape[1],3),dtype=numpy.uint8)
            rgb[:,:,0]=red
            rgb[:,:,1]=green
            rgb[:,:,2]=blue
            im = Image.fromarray(rgb,mode='RGB')
        if shrink  is not None: 
            # scale in order to make the largest dimension equal to 'shrink' 
            shrink = int(shrink) 
            (h,w) = im.size
#FIXME: downsample the volume (with the GPU, if available) before converting to images, it will save a lot of time, conversion to Image and to RGB is slow
            if h>shrink or w>shrink: 
                if h>w: 
                    im = im.resize( (shrink, int(shrink*w/(1.0*h))) )
                else: 
                    im = im.resize( (int(shrink*h/(1.0*w)),shrink) )
        if rotate is not None: 
            im = im.rotate(rotate)
        return im
        

    # display
    def display_in_browser(self,axis=None,shrink=None,rotate=None,subsample_slices=None,scales=None): 
        self.display(axis,shrink,rotate,subsample_slices,scales,open_browser=True)

    def display(self,axis=None,shrink=None,rotate=None,subsample_slices=None,scales=None,open_browser=None):
        if axis is None:
            axis = self._axis
        if shrink  is None:
            shrink = self._shrink
        if rotate  is None: 
            rotate = self._rotate
        if  subsample_slices  is None: 
            subsample_slices = self._subsample_slices
        if subsample_slices is None: 
            subsample_slices=1
        if scales  is None: 
            scales = self._scales 
        if open_browser  is None: 
            open_browser = self._open_browser
        if open_browser  is None: 
            open_browser = False 
        D = DisplayNode()
        images = []
        n=0
        
        N_slices = 0
        for j in range(len(self.volumes)): 
            N_slices += self.get_shape(j)[axis] 
        self._progress_bar = ProgressBar(height='6px', width='100%%', background_color=C.LIGHT_GRAY, foreground_color=C.GRAY)
        
        for j in range(len(self.volumes)): 
            if scales  is None: 
                scale = 255/(self.get_data(j).max()+1e-9)
            else: 
                scale = scales[j]*255/(self.get_data(j).max()+1e-9)
            images_inner = []
            for i in range(0,self.get_shape(j)[axis],subsample_slices): 
                im = self.export_image(j,i,axis,normalise=True,scale=scale,shrink=shrink,rotate=rotate)
                images_inner.append(im) 
                n+=1
                self._progress_bar.set_percentage(n*100.0/N_slices*subsample_slices) 
            images.append(images_inner) 
        return D.display('tipix', images, open_browser)   

    def _repr_html_(self): 
        return self.display()._repr_html_()
        






class MultipleVolumesNiftyPy(): 
    def __init__(self,volumes,axis=0,open_browser=None): 
        self.volumes = volumes 
        self._axis = axis
        self._open_browser = open_browser
        self._progress_bar = ProgressBar(height='6px', width='100%%', background_color=C.LIGHT_GRAY, foreground_color=C.GRAY)

    def get_data(self,volume_index): 
        volume = self.volumes[volume_index] 
        if isinstance(volume,numpy.ndarray): 
            return volume
        else:
            # Image3D
            return volume.data
        
    def get_shape(self,volume_index): 
        return self.volumes[volume_index].shape
        
    def _resample_volume(self): 
        pass 

    # display
    def display_in_browser(self,axis=None,max_size=200): 
        self.display(axis,max_size,open_browser=True)

    def display(self,axis=None, max_size=256, open_browser=None):
        if axis is None:
            axis = self._axis
        if open_browser  is None: 
            open_browser = self._open_browser
        if open_browser  is None: 
            open_browser = False 
        D = DisplayNode() 
        
        self._progress_bar = ProgressBar(height='6px', width='100%%', background_color=C.LIGHT_GRAY, foreground_color=C.GRAY)
        
        volume = self.volumes[0]    #FIXME: optionally use other grid
        # make grid of regularly-spaced points 
        box_min = volume.get_world_grid_min() 
        box_max = volume.get_world_grid_max() 
        span = box_max-box_min 
        max_span = span.max() 
        n_points = numpy.uint32(span / max_span * max_size) 
        grid = volume.get_world_grid(n_points) 
        n=0
        images = [] 
        for index in range(len(self.volumes)): 
                volume = self.volumes[index] 
                resampled = volume.compute_resample_on_grid(grid).data
                self._resampled = resampled
                sequence = []
                for slice_index in range(n_points[axis]): 
                    if axis == 0:
                        a = numpy.float32(resampled[slice_index,:,:].reshape((resampled.shape[1],resampled.shape[2]))) 
                    elif axis == 1: 
                        a = numpy.float32(resampled[:,slice_index,:].reshape((resampled.shape[0],resampled.shape[2])))
                    else: 
                        a = numpy.float32(resampled[:,:,slice_index].reshape((resampled.shape[0],resampled.shape[1])))
                    lookup_table = volume.get_lookup_table()
                    im = self.__array_to_im( a, lookup_table )
                    sequence.append(im.rotate(90))  #FIXME: make optional 
                    n+=1
                    self._progress_bar.set_percentage(n*100.0/(len(self.volumes)*max_size)) 
                images.append(sequence) 
        if len(images)==1: 
            return D.display('tipix', images[0], open_browser) 
        else:  
            return D.display('tipix', images, open_browser)   

    def __array_to_im(self, a, lookup_table): 
        if lookup_table  is not None: 
            red,green,blue,alpha = lookup_table.convert_ndarray_to_rgba(a)
            rgb = numpy.zeros((a.shape[0],a.shape[1],3),dtype=numpy.uint8)
            rgb[:,:,0]=red
            rgb[:,:,1]=green
            rgb[:,:,2]=blue
            im = Image.fromarray(rgb,mode='RGB') 
        else: 
            im = Image.fromarray(a).convert("RGB") 
        return im 
        
    def _repr_html_(self): 
        return self.display()._repr_html_()








def deg_to_rad(deg):
    return deg*numpy.pi/180.0

def rad_to_deg(rad):
    return rad*180.0/numpy.pi




try: 
    from NiftyPy.NiftyRec import SPECT_project_parallelholes as projection
except: 
    has_NiftyPy = False
    print "Please install NiftyPy"
else: 
    has_NiftyPy = True





class SPECT_Projection(): 
    """SPECT projection object. """
    def __init__(self, data): 
        self.data = data        
    
    def get_data(self):
        """Returns the raw projection data (note that is can be accessed also as self.data ). """
        return data

    def save_to_file(self, filename): 
        h5f = h5py.File(filename, 'w')
        h5f.create_dataset('data', data=self.data)
        #h5f.create_dataset('size_x', data=size_x)
        #h5f.create_dataset('size_y', data=size_y)
        h5f.close()

    def get_integral(self): 
        return self.data.sum() 

    def to_image(self,data,index=0,scale=None,absolute_scale=False): 
        from PIL import Image
        a = numpy.float32(data[:,:,index].reshape((data.shape[0],data.shape[1])))   
        if scale is None:
            a = 255.0*(a)/(a.max()+1e-12)
        else: 
            if absolute_scale: 
                a = scale*(a) 
            else: 
                a = scale*255.0*(a)/(a.max()+1e-12)
        return Image.fromarray(a).convert("RGB")
        
    def display_in_browser(self,axial=True,azimuthal=False,index=0,scale=None): 
        self.display(axial=axial,azimuthal=azimuthal,index=index,scale=scale,open_browser=True)

    def display(self,scale=None,open_browser=False): 
        data = self.data
        d = DisplayNode()
        images = []
        progress_bar = ProgressBar(height='6px', width='100%%', background_color=C.LIGHT_GRAY, foreground_color=C.GRAY) 
        if scale is not None:
            scale = scale*255.0/(data.max()+1e-12)
        else: 
            scale = 255.0/(data.max()+1e-12) 
        N_projections = self.data.shape[2]
        N_x = self.data.shape[0]
        N_y = self.data.shape[1]
        print "SPECT Projection   [N_projections: %d   N_x: %d   N_y: %d]"%(N_projections,N_x,N_y)
        for i in range( N_projections ): 
                images.append( self.to_image(data,i,scale=scale,absolute_scale=True) ) 
                progress_bar.set_percentage(i*100.0/N_projections)                         
        progress_bar.set_percentage(100.0)
        return d.display('tipix', images, open_browser)
        
    def _repr_html_(self): 
        return self.display()._repr_html_()
        

class VolumeRenderer(): 
    def __init__(self, volume, theta_min_deg=0.0, theta_max_deg=360.0, n_positions=180, truncate_negative=False, psf=None, attenuation=None, scale=1.0): 
        self.volume  = volume 
        self.psf = psf
        self.attenuation = attenuation
        self.use_gpu = is_gpu_enabled() 
        self.theta_min_deg = theta_min_deg 
        self.theta_max_deg = theta_max_deg 
        self.n_positions = n_positions
        self.truncate_negative = truncate_negative
        self.scale=scale

    def __make_cameras(self,axis,direction): 
        #self.cameras = numpy.float32(numpy.linspace(deg_to_rad(self.theta_min_deg),deg_to_rad(self.theta_max_deg),self.n_positions).reshape((self.n_positions,1))) 
        self.cameras = numpy.zeros((self.n_positions,3),dtype=numpy.float32)
        self.cameras[:,0] = self.cameras[:,0] + deg_to_rad(axis[0])
        self.cameras[:,1] = self.cameras[:,0] + deg_to_rad(axis[1])
        self.cameras[:,2] = self.cameras[:,0] + deg_to_rad(axis[2])
        self.cameras[:,direction] = numpy.linspace(deg_to_rad(self.theta_min_deg),deg_to_rad(self.theta_max_deg),self.n_positions)
        
    def render(self, axis=[90,0,0], direction=0, max_n_points=None): 
        if hasattr(self.volume,'compute_resample_on_grid'):   #i.e. if it is a Image3 
            volume = self.volume.copy()
            # make grid of regularly-spaced points
            if max_n_points is None: 
                max_n_points = 256
            box_min = volume.get_world_grid_min() 
            box_max = volume.get_world_grid_max() 
            span = box_max-box_min 
            max_span = span.max() 
            n_points = numpy.uint32(span / max_span * max_n_points) 
            grid = volume.get_world_grid(n_points) 
            volume.compute_resample_on_grid(grid)
            volume = numpy.float32(volume.data)
        else: 
            volume = self.volume
        self.__make_cameras(axis, direction)
        if has_NiftyPy: 
            proj_data = projection(volume, self.cameras, self.attenuation, self.psf, 0.0, 0.0, self.use_gpu, self.truncate_negative)
        else: 
            raise InstallationError("NiftyPy not installed, please install to execute render(). ")
        self.__proj = SPECT_Projection(proj_data)
        return self.__proj  #FIXME: memoize projection (use new style objects - properties)

    def display(self): 
        r = self.render()
        return r.display(scale=self.scale) 

    def _repr_html_(self):
        return self.display()





H_graph = """
<div id="__id__"></div>
<style type="text/css">
path.link {
  fill: none;
  stroke: #666;
  stroke-width: 1.5px;
}

marker#t0 {
  fill: green;
}

path.link.t0 {
  stroke: green;
}

path.link.t2 {
  stroke-dasharray: 0,2 1;
}

circle {
  fill: #ccc;
  stroke: #333;
  stroke-width: 1.5px;
}

text {
  font: 10px sans-serif;
  pointer-events: none;
}

text.shadow {
  stroke: #fff;
  stroke-width: 3px;
  stroke-opacity: .8;
}
</style>
"""

J_graph = """  
<script type="text/Javascript">
require.config({paths: {d3: "http://d3js.org/d3.v3.min"}});
require(["d3"], function(d3) {

var width = 800;
var height = 500;

var graph = __graph_data__; 

var nodes = graph['nodes']; 
var links = graph['links']; 

var nodes2 = {};
var links2 = []; 

for (var i=0; i<nodes.length; i++) {
    node = nodes[i]; 
    nodes2[node['name']] = node; 
};

for (var i=0; i<links.length; i++) {
    links2[i] = {'source':nodes2[links[i]['source']], 'target':nodes2[links[i]['target']], 'type':links[i]['type'],}; 
};


var force = d3.layout.force()
    .nodes(d3.values(nodes2))
    .links(links2)
    .size([width, height])
    .linkDistance(60)
    .charge(-300)
    .on("tick", tick)
    .start();

var svg = d3.select("#__id__").append("svg:svg")
    .attr("width", width)
    .attr("height", height);

// Per-type markers, as they don't inherit styles.
svg.append("svg:defs").selectAll("marker")
    .data(["t0", "t1", "t2"])
  .enter().append("svg:marker")
    .attr("id", String)
    .attr("viewBox", "0 -5 10 10")
    .attr("refX", 20)
    .attr("refY", -1.5)
    .attr("markerWidth", 6)
    .attr("markerHeight", 6)
    .attr("orient", "auto")
  .append("svg:path")
    .attr("d", "M0,-5L10,0L0,5");

var path = svg.append("svg:g").selectAll("path")
    .data(force.links())
  .enter().append("svg:path")
    .attr("class", function(d) { return "link " + d.type; })
    .attr("marker-end", function(d) { return "url(#" + d.type + ")"; });

colors  = ['#fff','#ccc','#ee4','#dddddd','#eeeeee','#ffffc0']; 
strokes = ['#333','#333','#333','#dddddd','#dddddd','#dddddd']; 

var circle = svg.append("svg:g").selectAll("circle")
    .data(force.nodes())
  .enter().append("svg:circle")
    .attr("r", 10)
    .call(force.drag)
    .style('fill', function(d){return colors[d.type];})
    .style('stroke', function(d){return strokes[d.type];});

var text = svg.append("svg:g").selectAll("g")
    .data(force.nodes())
  .enter().append("svg:g");

// A copy of the text with a thick white stroke for legibility.
text.append("svg:text")
    .attr("x", 12)
    .attr("y", ".31em")
    .attr("class", "shadow")
    .text(function(d) { return d.name; });

text.append("svg:text")
    .attr("x", 12)
    .attr("y", ".31em")
    .text(function(d) { return d.name; });


// Use elliptical arc path segments to doubly-encode directionality.
function tick() {
  path.attr("d", function(d) {
    var dx = d.target.x - d.source.x,
        dy = d.target.y - d.source.y,
        dr = Math.sqrt(dx * dx + dy * dy);
    return "M" + d.source.x + "," + d.source.y + "A" + dr + "," + dr + " 0 0,1 " + d.target.x + "," + d.target.y;
  });

  circle.attr("transform", function(d) {
    return "translate(" + d.x + "," + d.y + ")";
  });

  text.attr("transform", function(d) {
    return "translate(" + d.x + "," + d.y + ")";
  });
}

});
</script>
"""


import json 

class Graph(): 
    def __init__(self, graph): 
        self.set_graph(graph)

    def set_graph(self,g):
        self.graph = g

    def get_html(self): 
        div_id = "graph_"+str(uuid.uuid4())
        H = H_graph.replace("__id__",div_id)
        J = J_graph.replace("__id__",div_id)
        J = J.replace("__graph_data__",json.dumps(self.graph))
        return H+J

    def _repr_html_(self): 
        graph = HTML( self.get_html() )
        return graph._repr_html_() 

    


## ipy_table 
#FIXME: perhaps move this somewhere else

try: 
    import ipy_table 
    has_ipy_table = True
except: 
    print "Please install ipy_table (e.g. 'easy_install ipy_table') to enable ipython notebook tables. "
    ipy_table = None 
    has_ipy_table = False




## svg_write
#FIXME: perhaps move this somewhere else

try: 
    import svgwrite
    has_svgwrite = True
except: 
    print "Please install svgwrite (e.g. 'easy_install svgwrite') to enable svg visualisations. "
    svgwrite = None
    has_svgwrite = False




        

    




# occiput 
# Stefano Pedemonte
# Aalto University, School of Science, Helsinki
# Oct 2013, Helsinki 
# Harvard University, Martinos Center for Biomedical Imaging 
# Boston, MA, USA



import nibabel
from occiput.Core.Conversion import nipy_to_occiput
import numpy 



# FIXME: move this class to occiput.Visualization 

class ColorLookupTable():
    def __init__(self): 
        self._by_index = {}
        self._by_name  = {} 
        
    def n_entries(self): 
        return len(self._by_index)

    def load_from_file_freesurfer(self, filename): 
        with open(filename,'r') as fid: 
            F = fid.read() 
            F = F.split('\r\n') 
        self._by_index = {}
        self._by_name  = {} 
        for line in F: 
            if not line=='': 
                index, name, r, g, b, a = line.split() 
                r = numpy.uint8(r)
                g = numpy.uint8(g)
                b = numpy.uint8(b)
                a = numpy.uint8(a)
                self._by_index[str(index)] = {'name':name,'r':r,'g':g,'b':b,'a':a}
                self._by_name[str(name)]   = {'index':index,'r':r,'g':g,'b':b,'a':a}

    def index_to_rgba(self,index):
        entry = self._by_index[str(index)]
        return entry['r'],entry['g'],entry['b'],entry['a']

    def name_to_rgba(self,name): 
        entry = self._by_name[str(name)]
        return entry['r'],entry['g'],entry['b'],entry['a']
    
    def index_to_name(self,index): 
        entry = self._by_index[str(index)] 
        return entry['name']

    def has_index(self,index): 
        return str(index) in self._by_index.keys() 

    def default_rgba(self): 
        if self.has_index(0): 
            return self.index_to_rgba(0)
        else: 
            return numpy.uint8(0), numpy.uint8(0), numpy.uint8(0), numpy.uint8(0)
    
    def convert_ndarray_to_rgba(self, array): 
        unique = numpy.unique(numpy.uint32( array ))
        R = numpy.zeros(array.shape)
        G = numpy.zeros(array.shape)
        B = numpy.zeros(array.shape)
        A = numpy.zeros(array.shape)
        for index in unique: 
            if self.has_index(index): 
                r,g,b,a = self.index_to_rgba(index) 
            else: 
                r,g,b,a = self.default_rgba() 
            I = numpy.where(array==index)
            R[I] = r
            G[I] = g
            B[I] = b
            A[I] = a
        return R,G,B,A




def load_freesurfer_lut_file(filename): 
    lut = ColorLookupTable()
    lut.load_from_file_freesurfer(filename)
    return lut 






# occiput 
# Stefano Pedemonte
# Aalto University, School of Science, Helsinki
# Oct 2014, Helsinki 
# Harvard University, Martinos Center for Biomedical Imaging 
# Boston, MA, USA
# April 2014
# Nov. 2015 


# Import an interfile volume as an Image3D and export. 


from occiput.Core import Image3D, Transform_Identity, Transform_6DOF, Transform_Affine, Transform_Scale
from interfile import Interfile
import h5py
import os
from numpy import isscalar, linspace, int32, uint32, ones, zeros, pi, sqrt, float32, float64, where, ndarray, nan
from numpy import inf, asarray, concatenate, fromfile, maximum, exp, asfortranarray, fliplr, transpose  


__all__ = ["import_interfile_volume","export_interfile_volume"]

def import_interfile_volume_data(headerfile='', datafile=''):  #FIXME: this should be in the Interfile package
        F = Interfile.load(headerfile)
        if F.has_key('matrix size[1]'): 
            Nx    = F['matrix size[1]']['value']
            Ny    = F['matrix size[2]']['value']
            Nz    = F['matrix size[3]']['value']        
        else:
            Nx    = F['matrix size [1]']['value']
            Ny    = F['matrix size [2]']['value']
            Nz    = F['matrix size [3]']['value']    
        if datafile == '': 
            datafile1 = headerfile.replace(headerfile.split(os.sep)[-1],F['name of data file']['value'])
            datafile2 = headerfile.replace('.v.hdr','.v') 
            datafile2 = datafile2.replace('.h33','.v')
            datafile3 = headerfile.replace('.h33','.v')
            try: 
                data = fromfile(datafile1,dtype=float32)
            except: 
                try: 
                    data = fromfile(datafile2,dtype=float32)
                except: 
                    try: 
                        data = fromfile(datafile3,dtype=float32)
                    except: 
                        print "Data file not found."
        else: 
            data = fromfile(datafile,dtype=float32)
        data = data.reshape([Nz,Ny,Nx])
        data = asfortranarray(data.swapaxes(0,2))
        data = transpose(data,[1,0,2])
        data = data[::-1,:,:]
        return data    

def import_interfile_volume(headerfile='', datafile=''): 
        # Load ndarray data 
        data = import_interfile_volume_data(headerfile, datafile) 
        # Load other information - e.g. pixels size 
        F = Interfile.load(headerfile)
        if F.has_key('scale factor (mm/pixel) [1]'): 
            pixsize_x = F['scale factor (mm/pixel) [1]']['value']
            pixsize_y = F['scale factor (mm/pixel) [2]']['value']
            pixsize_z = F['scale factor (mm/pixel) [3]']['value']
        # Create Image3D 
        T_pix_to_world = Transform_Scale(int32([pixsize_x,pixsize_y,pixsize_z]), map_from='pixels_PET', map_to='world') 
        volume = Image3D(data=data, affine=T_pix_to_world, space='world')  
        return volume


def export_interfile_volume(data_file_name, data): 
        data = data[::-1,:,:]
        data = transpose(data,[1,0,2])
        data = data.swapaxes(0,2)
        data = asarray(data, dtype=float32, order='C')
        data.tofile(data_file_name)








# occiput  
# Stefano Pedemonte
# Harvard University, Martinos Center for Biomedical Imaging 
# Apr. 2014, Boston, MA

import numpy
from occiput.Core import Image3D, Grid3D, Transform_6DOF
from ilang.Models import Model 




class SSD_ilang( Model ):
    variables = {'source':'continuous','target':'continuous','transformation':'continuous','sigma':'continuous'} 
    dependencies = [['source','target','directed'],['transformation','target','directed'],['sigma','target','directed']]

    def __init__( self, name=None ):
        if name == None:  
            name = self.__class__.__name__
        Model.__init__(self, name) 
    
    def log_conditional_probability_transformation(self,transformation): 
        source = self.get_value('source') 
        target = self.get_value('target') 
        sigma  = self.get_value('sigma')
        log_p = 0.0
        return log_p

    def log_conditional_probability_gradient_transformation(self,transformation): 
        source = self.get_value('source') 
        target = self.get_value('target') 
        sigma  = self.get_value('sigma')
        gradient = numpy.asarray([0.0,0.0,0.0,0.0,0.0,0.0]) 
        return gradient 
        
    def sample_conditional_probability_target(self): 
        return 0    

    def init(self): 
        pass 














#!/usr/bin/python
# -*- coding: UTF-8 -*-
          
# occiput 
# Stefano Pedemonte
# Aalto University, School of Science, Helsinki
# Oct 2013, Helsinki 
# Harvard University, Martinos Center for Biomedical Imaging 
# Boston, MA, USA


from occiput.Visualization.Visualization import ProgressBar
from occiput.DataSources.FileSources.LookupTable import load_freesurfer_lut_file

import subprocess
import logging
import sys
import inspect
import os


class Downloader_HTTP(): 
    def __init__(self): 
        self._filename = 'unknown' 
        self._output = None 
        self._progress_bar = ProgressBar() 
        self._progress_bar.set_percentage(0) 
        self._verbose = False 

    def set_verbose(self, verbose): 
        self._verbose = verbose 

    def _set_percentage(self,percentage): 
        self._progress_bar.set_percentage(percentage) 

    def download(self, url, output=None, overwrite=False): 
        # Check if output exists: 
        if not overwrite: 
            if output!=None: 
                if os.path.exists(output): 
                    if self._verbose: 
                        print "File",output,"exists, not dowloading." 
                    self._set_percentage(100) 
                    return output 
        self._set_percentage(0) 
        self._output = output 
        if output==None: 
            args = ['wget',url,'--continue']
        else: 
            args = ['wget',url,'--continue','-O',output]
        try:
            pipe = subprocess.Popen(args, bufsize = 0,
                shell = False,
                stdout = None, # no redirection, child use parent's stdout
                stderr = subprocess.PIPE) # redirection stderr, create a new pipe, from which later we will read
        except Exception as e:
            #inspect.stack()[1][3] will get caller function name
            logging.error(inspect.stack()[1][3] + ' error: ' + str(e))
            return False
        while 1:
            s = pipe.stderr.readline()
            if s:
                p = self._strip_percentage(s)
                if p!=None: 
                    self._set_percentage(p) 
                if self._verbose:  
                    print s
                name = self._strip_filename(s) 
                if name!=None: 
                    self._set_filename(name)
                #sys.stdout.write(s)
            if pipe.returncode is None:
                code = pipe.poll()
            else:
                break
        if not 0 == pipe.returncode:
            self._set_percentage(0) 
            return False
        self._set_percentage(100)
        if output: 
            return output
        else: 
            return self._filename

    def _strip_percentage(self,s):
        p = s[s.find('%')-2:s.find('%')].strip(' ') 
        try: 
            percentage = int(p)
        except: 
            return None
        else: 
            return percentage
 
    def _strip_filename(self,s): 
        if s.find("Saving to")!=-1: 
            name = s.strip("Saving to").strip("\n").strip("“").strip("'").strip("`")
            if self._verbose: 
                print "Detected name: ",name
            return name 

    def _set_filename(self,name): 
        self._filename = name 

class Dropbox(Downloader_HTTP): 
    pass 
    

def download_Dropbox(url, output=None, overwrite=False, verbose=False): 
    D = Dropbox()
    D.set_verbose(verbose)
    return D.download(url, output, overwrite) 





# occiput.io 
# Stefano Pedemonte
# Harvard University, Martinos Center for Biomedical Imaging 
# 2013 - 2015
# Boston, MA


# Interface to the ray-tracer for projection and back-projection. 
# The interface to the ray-tracer for scatter simulation is in another file. 


__all__ = ["DEFAULT_PROJECTION_PARAMETERS","DEFAULT_BACKPROJECTION_PARAMETERS","ProjectionParameters","BackprojectionParameters", 
           "PET_project_compressed","PET_backproject_compressed","has_NiftyPy"]

from occiput.Core.NiftyPy_wrap import PET_project_compressed, PET_backproject_compressed, has_NiftyPy


DEFAULT_PROJECTION_PARAMETERS  =  {
                         "N_samples":                256, 
                         "sample_step":              0.8, 
                         "background_activity":      0.0, 
                         "background_attenuation":   0.0, 
                         "truncate_negative_values": 0,  
                         "gpu_acceleration":         1,      
                         "direction":                7,      # affects performance only, between 1 and 6; 2 and 5 are normally the best value
                         "block_size":               512 }   # affects performance only

DEFAULT_BACKPROJECTION_PARAMETERS  =  {
                         "N_samples":                256, 
                         "sample_step":              0.8, 
                         "background_activity":      0.0, 
                         "background_attenuation":   0.0, 
                         "truncate_negative_values": 0,  
                         "gpu_acceleration":         1,      
                         "direction":                7,      # affects performance only, between 1 and 6; 2 and 5 are normally the best value
                         "block_size":               512 }   # affects performance only



class ProjectionParameters(): 
    """Data structure containing the parameters of a projector for PET. """
    default_parameters = DEFAULT_PROJECTION_PARAMETERS
    def __init__(self,parameters=None): 
        self._initialised = 0
        self.name = "Unknown binning name"
        if parameters is None: 
            self.load_from_dictionary(self.default_parameters)  
        elif type(parameters) == dict:  
            self.load_from_dictionary(parameters)        
        elif type(parameters) in [list, tuple]: 
            if len(parameters) == len(self.default_parameters.keys()): 
                self.N_samples                = parameters[0] 
                self.sample_step              = parameters[1]
                self.background_activity      = parameters[2]
                self.background_attenuation   = parameters[3] 
                self.truncate_negative_values = parameters[4]
                self.gpu_acceleration         = parameters[5]
                self.direction                = parameters[6]
                self.block_size               = parameters[7]
            else: 
                raise UnknownParameter('Parameter %s specified for ProjectionParameters is not compatible. '%str(parameters)) 
        else: 
            raise UnknownParameter('Parameter %s specified for ProjectionParameters is not compatible. '%str(parameters)) 
            
    def load_from_dictionary(self,dictionary):
        self.N_samples                = dictionary['N_samples']                 # Number of samples along a line when computing line integrals
        self.sample_step              = dictionary['sample_step']               # distance between consecutive points along a line when computing line integrals (this is in the same unit measure as the size of the imaging volume (activity_size and attenuation_size))
        self.background_activity      = dictionary['background_activity']       # Activity in voxels outside of the imaging volume
        self.background_attenuation   = dictionary['background_attenuation']    # Attenuation in voxels outside of the imaging volume 
        self.truncate_negative_values = dictionary['truncate_negative_values']  # If set to 1, eventual negative values obtained when projecting are set to 0
                                                                                # (This is meant to remove eventual unwanted small negative values due to FFT-based smoothing within the projection algorithm 
                                                                                # - note that numerical errors may produce small negative numbers when doing FFT-IFFT even if the function is all positive )
        self.gpu_acceleration         = dictionary['gpu_acceleration']          # Whether to use GPU acceleration (if available) or not 
        self.direction                = dictionary['direction']                 # Direction parameter for the projector and b-proj., it affects performance only; 2 and 5 are normally the best values. 
        self.block_size               = dictionary['block_size']                # Number of blocks per thread for the GPU accelerated projector and b-proj. It affects performance only. Typical values: 256, 512, 768
        


class BackprojectionParameters(ProjectionParameters): 
    """Data structure containing the parameters of a projector for PET. """
    default_parameters = DEFAULT_BACKPROJECTION_PARAMETERS
    # Note: this seems unfinished, but it inherits all that is necessary from ProjectionParameters


      




# occiput   
# Stefano Pedemonte
# Harvard University, Martinos Center for Biomedical Imaging
# Jan 2014, Boston, MA, USA 


__all__ = ['RigidTransformationSSD']

from ilang.Models import Model 
from ilang.Graphs import ProbabilisticGraphicalModel 
from ilang.Samplers import Sampler

from occiput.Core import Image3D 
from occiput.Visualization import MultipleVolumes

try: 
    from NiftyPy.NiftyReg import resample_image_rigid
    from NiftyPy.NiftyReg import deriv_intensity_wrt_space_rigid
    from NiftyPy.NiftyReg import deriv_intensity_wrt_transformation_rigid
    from NiftyPy.NiftyReg import deriv_ssd_wrt_transformation_rigid
    from NiftyPy.NiftyReg import gaussian_smoothing
except: 
    has_NiftyPy = False
    print "Please install NiftyPy"
else: 
    has_NiftyPy = True 

import numpy





class ModelRigidSSD(Model): 
    variables          = {'source':'continuous','target':'continuous','transformation':'continuous'} 
    dependencies       = [['source','target','directed'],['transformation','target','directed']] 
    preferred_samplers = {'transformation':['QuasiNewton_L_BFGS_B']}

    # init
    def init(self): 
        pass 
         
    # expose to sampler: 
    def log_conditional_probability_transformation(self,T): 
        source    = self.get_value('source')
        target    = self.get_value('target')
        #return -.5*numpy.dot(numpy.dot((x-mu),hessian),(x-mu).T)
        FIXME
        return 0

    def log_conditional_probability_gradient_transformation(self,T): 
        source    = self.get_value('source')
        target    = self.get_value('target')      
        #return -.5*numpy.dot((x-mu),hessian+hessian.T) 
        # FIXME
        return numpy.random.rand(6)*4/30000






class RigidTransformationSSD(): 
    def __init__(self,target,source): 
        self.target = target 
        self.source = source 
        self._resampled_source_cache = None 
        self._resampled_source_need_update = True
        self._contruct_ilang_model()
        self._set_transformation(numpy.zeros((1,6))) 
        
    def _contruct_ilang_model(self): 
        self.ilang_model   = ModelRigidSSD('RigidSSD')
        self.graph         = ProbabilisticGraphicalModel(['target','source','transformation']) 
        self.graph.set_nodes_given(['target','source'],True) 
        self.graph.set_node_value('source',self.source.get_data())
        self.graph.set_node_value('target',self.target.get_data())
        self.graph.set_node_value('transformation',numpy.zeros((1,6)))
        self.graph.add_dependence(self.ilang_model,{'target':'target','source':'source','transformation':'transformation'}) 
        self.sampler       = Sampler(self.graph) 

    def _set_transformation(self,transformation):
        self.graph.set_node_value('transformation',transformation) 
        self._resampled_source_need_update = True 
    
    def _get_transformation(self): 
        return self.graph.get_node_value('transformation')
      
    def estimate_transformation(self,method=None,iterations=30000,trace=True,parameters=None, display_progress=True): 
        if method!=None: 
            self.sampler.set_node_sampling_method_manual('transformation',method)
        else: 
            self.sampler.set_node_sampling_method_auto('transformation')
        last_sample = self.sampler.sample_node('transformation',nsamples=iterations,trace=trace) 
        # This is not necessary, but it triggers the cache of the resampled source image: 
        self._set_transformation(last_sample) 
        return last_sample 

    def resample_in_target_space(self,image): 
        # FIXME 
        #print "Resampling .."
        transformation = self._get_transformation() 
        return image

    def _get_resampled_source(self): 
        #FIXME: remove the next few lines
        if self.transformation.sum() != 0: 
            return self.target
        
        if self._resampled_source_need_update: 
            resampled_source = self.resample_in_target_space(self.source)
            self._resampled_source_cache = resampled_source
            self._resampled_source_need_update = False 
        else: 
            resampled_source = self._resampled_source_cache
        return resampled_source 

    def display_in_browser(self,axis=0,shrink=256,rotate=90,subsample_slices=4,scale_factors=None):
        self.display(axis,shrink,rotate,scale_factors,open_browser=True)
        
    def display(self,axis=0,shrink=256,rotate=90,subsample_slices=4,scale_factors=None,open_browser=False): 
        resampled = self.resampled_source
        D = MultipleVolumes([resampled,self.target], axis, shrink, rotate, subsample_slices, scale_factors, open_browser)
        return D.display() 

    def _repr_html_(self): 
        return self.display()._repr_html_()

    transformation   = property(_get_transformation, _set_transformation, None)
    resampled_source = property(_get_resampled_source, None, None)





# occiput
# Stefano Pedemonte
# Harvard University / MGH
# Boston, Jan 2015


from .. import occiput

import unittest


class Test_Reconstruction_PET(unittest.TestCase): 
    """Sequence of tests for tomographic reconstruction - Positron Emission Tomography """   
    def setUp(self):
        pass 

    def test_projection_wrapper(self):
        """Test the Python wrapper for the projection algorithm. """
        number = 0.1
        descriptor = [  {'name':'input',  'type':'int', 'value':number},
                        {'name':'output', 'type':'int', 'value':None },  ]
        r = c_python.call_c_function( self.lib.echo, descriptor ) 
        self.assertTrue(r.output == number)

    def test_projection(self):
        """Test the projection algorithm. """
        A = 0
        B = 0
        self.assertTrue(A == B)


class Test_Reconstruction_SPECT(unittest.TestCase): 
    def setUp(self):
        pass 
        
class Test_Reconstruction_CT(unittest.TestCase): 
    def setUp(self): 
        pass 



if __name__ == '__main__':
    unittest.main()
    





try: 
    from Occiput_Interface_Biograph_mMR import Biograph_mMR_Physiology
except: 
    Biograph_mMR_Physiology = None 

try: 
    from Occiput_Interface_Brain_PET import Brain_PET_Physiology
except: 
    Brain_PET_Physiology = None 











# occiput.io 
# Stefano Pedemonte
# Harvard University, Martinos Center for Biomedical Imaging 
# 2013 - 2015
# Boston, MA


# Occiput.io does not like the concepts of sinogram-span and meshing. 
# These concepts have been utilized in the past in order to reduce the computational 
# complexity of PET reconstruction. However sinogram meshing makes the emission data 
# complicated and the complicates the reconstruction software. 
# Occiput.io can do without. It uses the GPU acceleration to avoid meshing altogether and 
# it also provides a technique similar to meshing to further speed-up the reconstruction. 
# This is full automated, it does not require the concept of span and Michelogram.  
# For compatibility, however, occiput can deal with the conventional meshing technique. 
# The code for conventional meshing is here. 



from occiput.Visualization import svgwrite, has_svgwrite
from numpy import int32, ones 


class Michelogram: 
    def __init__(self, n_rings=64, span=11, max_ring_difference=60): 
        self.n_rings = n_rings
        self.span = span
        self.max_ring_difference = max_ring_difference 
        self.n_sinograms = None 
        self.n_planes    = None 
        self._svg_string = None 
        self._make()
        self._make_svg()

    def _make(self): 
        if not self._verify_max_ring_difference(): 
            print "Michelogram: Maximum ring difference is not consistent with span. "
        n_sinograms_per_side = (self.max_ring_difference+1-self.span/2)/self.span 
        n_sinograms = n_sinograms_per_side * 2 + 1
        n_planes = zeros([n_sinograms_per_side+1,],dtype=int32)
        n_planes[0] = 2*n_rings-1
        n_planes[1] = n_planes[0] - self.span
        for i in range(n_sinograms_per_side-1): 
            n_planes[2+i]=n_planes[1+i]-2*span
        self.n_sinograms = n_sinograms
        self.n_planes    = n_planes

        self.michelogram_sinogram = -1*ones([n_rings,n_rings],order = 'F',dtype=int32)
        self.michelogram_plane    = -1*ones([n_rings,n_rings],order = 'F',dtype=int32)

        # -2- right sinograms  
        sinogram_index = 0
        for X in range(0,max_ring_diff,span): 
            if sinogram_index == 0: 
                plane_index_offset = -n_rings_small
            else: 
                plane_index_offset = 0
            self._fill_michelogram(X, -n_rings_small, sinogram_index, plane_index_offset)
            sinogram_index += 2

        # -3- left sinograms 
        sinogram_index = 1
        for Y in range(n_rings_large,max_ring_diff,span): 
            plane_index_offset = 0
            self._fill_michelogram(0, Y, sinogram_index, plane_index_offset)
            sinogram_index += 2
            
    def _verify_max_ring_difference(self): 
        return True
        #verify if max_ring_difference is consistent with span and n_rings (give warnings if not): 

    def _fill_michelogram(self, x_start, y_start, sinogram_index, plane_index_offset): 
        n_rings = self.n_rings
        n_rings_small = self.span/2
        n_rings_large = n_rings_small+1 
        for X in range(n_rings + n_rings_small): 
            # long line
            plane = 2*X + plane_index_offset
            for line_index in range(n_rings_large): 
                x = x_start + X - line_index
                y = y_start + X + line_index
                if x>=0 and x<n_rings and y>=0 and y<n_rings: 
                    self.michelogram_sinogram[x,y] = sinogram_index
                    self.michelogram_plane[x,y] = plane
            # short line
            plane = 2*X + 1 + plane_index_offset
            for line_index in range(n_rings_small): 
                x = x_start + X - line_index
                y = y_start + X + 1 + line_index
                if x>=0 and x<n_rings and y>=0 and y<n_rings: 
                    self.michelogram_sinogram[x,y] = sinogram_index
                    self.michelogram_plane[x,y] = plane 

    def _make_svg(self): 
        if not has_svgwrite: 
            self._svg_string = None 
            return self._svg_string 
        w = '100%'
        h = '100%'
        dwg = svgwrite.Drawing('michelogram.svg',size=(w,h), profile='full', debug=True)
        dwg.viewbox(width=100, height=100)

        # blank table 
        rect = dwg.add(dwg.rect(insert=(0, 0), size=(100, 100), rx=0.0, ry=0.0))
        rect.fill('grey',opacity=0.1).stroke('black',width=0.1,opacity=0.001)
        #for ... 
        #line = dwg.add(dwg.polyline([(10, 10), (10, 100), (100, 100), (100, 10), (10, 10)],stroke='black', fill='none'))

        # fill table with Michelogram data 

        self._svg_string = dwg.tostring() 
        return self._svg_string

    def _repr_svg_(self): 
        return self._svg_string    

    def display_image(self): 
        import pylab 
        pylab.imshow(self.michelogram_plane,interpolation='nearest',cmap='gray') 






# occiput  
# Stefano Pedemonte
# Harvard University, Martinos Center for Biomedical Imaging 
# Dec. 2014, Boston, MA 
# Apr. 2014, Boston, MA


# FIXME: perhaps get rid of this file and place code in occiput.Transformation (mostly), although as it is now, all imports from 
# NiftyPy are in one place (here). This is somewhat convenient to keep track of things. 


import occiput as __occiput
import numpy as __np
try:
    from NiftyPy.NiftyRec import INTERPOLATION_LINEAR, INTERPOLATION_POINT
    from NiftyPy.NiftyRec import TR_resample_grid as             __TR_resample_grid 
    from NiftyPy.NiftyRec import TR_grid_from_box_and_affine as  __TR_grid_from_box_and_affine 
    from NiftyPy.NiftyRec import TR_transform_grid as            __TR_transform_grid 
    from NiftyPy.NiftyRec import PET_project_compressed, PET_backproject_compressed 
    from NiftyPy.NiftyRec import PET_compress_projection, PET_uncompress_projection, PET_initialize_compression_structure
    from NiftyPy.NiftyRec import SPECT_project_parallelholes, SPECT_backproject_parallelholes 
    from NiftyPy.NiftyRec import gpu_set, gpu_reset, gpu_list, gpu_exists
except: 
    print "NiftyPy could not be loaded: it will not be possible to reconstruct the PET data. "
    has_NiftyPy = False
    PET_project_compressed = None
    PET_backproject_compressed = None
    SPECT_project_parallelholes = None
    SPECT_backproject_parallelholes = None 
    __TR_resample_grid = None
    __TR_grid_from_box_and_affine = None
    __TR_transform_grid = None
    raise 
else: 
    has_NiftyPy = True 



## Spatial Transformation 

def __make_grid(data,space,ndim):
    if ndim==3: 
        return __occiput.Core.Grid3D(data,space)
    else: 
        return __occiput.Core.GridND(data,space)


def transform_grid(grid, affine_from_grid): 
    # 1) verify if the spaces of the affine map and of the grid are compatible: 
    if not affine_from_grid.can_left_multiply(grid): 
        print "Affine transformation not compatible with grid. " 
        # FIXME: raise error, or warning, depending on a global setting 
    # 2) transform 
    transformed = __TR_transform_grid( grid.data, affine_from_grid.data ) 
    # 3) instantiate a new grid 
    grid = __make_grid(transformed, affine_from_grid.map_to, transformed.ndim-1)
    return grid



def grid_from_box_and_affine(min_coords, max_coords, n_points, affine=None, space="world"): 
    #FIXME: optionally use the affine to transform min_coords and max_coords 
    data = __TR_grid_from_box_and_affine(min_coords,max_coords,n_points)
    ndim = data.ndim-1
    grid = __make_grid(data, space, ndim)
    return grid



def resample_image_on_grid(image, grid, affine_grid_to_world=None, verify_mapping=True, background=0.0, use_gpu=1, interpolation_mode=None): 
    if verify_mapping:
        # check if the image, the grid and the affine mapping are compatible: 
        # 1) if affine_grid_to_world is not defined, verify if image and grid are compatible
        if affine_grid_to_world == None: 
            if not image.affine.can_inverse_left_multiply(grid): 
                print "grid and image are not compatible. "
                #FIXME: raise exception
                return       
        # 2) if affine_grid_to_world is defined, verify if image, grid and affine_grid_to_world are compatible 
        else: 
            # a) check mapping from grid to world 
            if not affine_grid_to_world.can_left_multiply(grid): 
                print "grid and affine_grid_to_world are not compatible. "
                #FIXME: raise exception
                return    
            # b) check mapping from image to world 
            if not image.affine.can_inverse_left_multiply(affine_grid_to_world):
                print "image and affine_grid_to_world are not compatible. "
                #FIXME: raise exception
                return
    # compute affine: 
    if affine_grid_to_world == None:
        affine = image.affine
    else: 
         affine = affine_grid_to_world.left_multiply(image.affine)
    # decide sampling mode
    if interpolation_mode is None: 
      if image.is_mask(): 
        interpolation_mode = INTERPOLATION_POINT
      else:
        interpolation_mode = INTERPOLATION_LINEAR 
    # resample: 
    resampled_data = __TR_resample_grid( __np.float32(image.data), __np.float32(grid.data), __np.float32(affine.data), background, use_gpu, interpolation_mode )
    return resampled_data 
    

    






# occiput  
# Stefano Pedemonte
# Harvard University, Martinos Center for Biomedical Imaging 
# Dec. 2014, Boston, MA 
# April. 2014, Boston, MA


import occiput as __occiput
import nibabel as __nibabel
import warnings as __warnings
with __warnings.catch_warnings():
    __warnings.simplefilter("ignore")
    import nipy as __nipy
    from nipy.io.nifti_ref import nifti2nipy as nifti_to_nipy



def nipy_to_occiput(nib): 
    if nib.ndim == 3: 
        im = __occiput.Core.Image3D(data=nib.get_data(), affine=nib.affine, space="world") 
    else: 
        im = __occiput.Core.ImageND(data=nib.get_data(), affine=nib.affine, space="world")   
    return im 



def nifti_to_occiput(nif): 
    nip = nifti_to_nipy(nif)
    return nipy_to_occiput(nip)



def occiput_to_nifti(occ): 
    nii = __nibabel.nifti1.Nifti1Image(occ.data,occ.affine.data) 
    return nii
    


def occiput_from_array(array):
    if array.ndim == 3: 
        im = __occiput.Core.Image3D(data=array, space="world") 
    else: 
        raise("Currently only conversion of 3D arrays is supported. ")
    return im 



# occiput - Inference Language 
# Stefano Pedemonte
# Aalto University, School of Science, Helsinki
# Oct 2013, Helsinki 
# Harvard University, Martinos Center for Biomedical Imaging, Boston
# Dec 2013, Boston, MA, USA 




# Print with 3 levels of verbosity 

__verbose = 1
def set_verbose_high(): 
    """Print everything - DEBUG mode"""
    global __verbose; __verbose = 2    
def set_verbose_medium(): 
    """Print runtime information"""
    global __verbose; __verbose = 1
def set_verbose_low(): 
    """Print only important messages"""
    global __verbose; __verbose = 0
def set_verbose_no_printing(): 
    """Do not print messages at all"""
    global __verbose; __verbose = -1        
def get_verbose_level():
    return __verbose
def print_debug(msg):
    """Use this for DEBUG Information"""
    if __verbose >= 2: 
        print msg
def print_runtime(msg):
    """Use this for messages useful at runtime"""
    if __verbose >= 1: 
        print msg
def print_important(msg):
    """Use this for important messages"""
    if __verbose >= 0: 
        print msg


# Other print options 

import contextlib as __contextlib
import numpy as __numpy
@__contextlib.contextmanager
def printoptions(*args, **kwargs):
    original = __numpy.get_printoptions()
    __numpy.set_printoptions(*args, **kwargs)
    yield 
    __numpy.set_printoptions(**original)



# GPU enables / disable

__use_gpu = True 
def enable_gpu():
    global __use_gpu; __use_gpu = True
def disable_gpu():
    global __use_gpu; __use_gpu = False 
def is_gpu_enabled():
    global __use_gpu
    return __use_gpu



# Default background of images

__background = 0.0
def set_default_background(bg):
    global __background; __background = bg
def get_default_background():
    global __background; 
    return __background






# occiput 
# Stefano Pedemonte 
# April 2014 
# Harvard University, Martinos Center for Biomedical Imaging 
# Boston, MA, USA 


import Colors
import Visualization
from Visualization import ProgressBar, VolumeRenderer, MultipleVolumes, MultipleVolumesNiftyPy, Graph, ipy_table, has_ipy_table, svgwrite, has_svgwrite
from ipynb import is_in_ipynb







# occiput 
# Stefano Pedemonte 
# April 2014 
# Harvard University, Martinos Center for Biomedical Imaging 
# Boston, MA, USA 



from . import Core
from .Core import Transform_Affine, Transform_Identity, Transform_6DOF, Transform_Scale, Transform_Translation, Transform_Rotation, Grid3D, Image3D, GridND, ImageND
from .Core import grid_from_box_and_affine
from .Core import nipy_to_occiput, nifti_to_occiput, occiput_from_array

from . import transformations 
from . import NiftyPy_wrap
from . import Conversion
from . import Errors
from . import Print 